from ..service import Operation


class DeleteFolder(Operation):
    """DeleteFolder EWS Operation attempts to delete a folder.
    Specifically this class is focused on deleting search folders.
    """

    RESULTS_KEY = 'Folders'

    def __init__(self, folder_id):
        """
        """
        self.folder_id = folder_id

    def soap(self):
        return self.M_NAMESPACE.DeleteFolder(
            self.M_NAMESPACE.FolderIds(
                self.T_NAMESPACE.FolderId(
                    Id=self.folder_id
                )
            ),
            DeleteType="MoveToDeletedItems"
        )



from .base import Base, etree, abc


class Operation(Base):
    """Operation class inherits from the Base class
    and defines namespaces, headers, and body
    of EWS Operation SOAP request
    """

    def get(self, exchange_version):
        self.__logger.info('Building SOAP request for {current}'.format(current=self.__class__.__name__))
        ENVELOPE = self.SOAP_MESSAGE_ELEMENT.Envelope
        HEADER = self.SOAP_NAMESPACE.Header
        self.envelope = ENVELOPE(
            HEADER(
                self.T_NAMESPACE.RequestedServerVersion(Version=exchange_version),
                self._impersonation_header()
            ),
            self.BODY_ELEMENT(
                self.soap()
            )
        )
        return etree.tostring(self.envelope)

    @abc.abstractmethod
    def soap(self):
        raise NotImplementedError



from os import access
from .exchangeversion import ExchangeVersion
from .core import Core
from .oauth2connector import OAuth2Connector


class AuthenticationProperties(type):

    def __set_initial_property_values(cls):
        if isinstance(cls._credentials, tuple):
            cls.domain = cls._credentials[0]
            if cls.impersonate_as != '':
                cls.auth_header = cls._credentials[0]
            else:
                cls.auth_header = None
            cls.ews_url = None
            cls.exchange_versions = None
            if cls.tenant_id and cls.client_id and cls.client_secret:
                if not cls.oauth2_authorization_type:
                    print('Please provide an OAuth2 Authorization Types before continuing')
                else:
                    try:
                        cls.access_token = getattr(OAuth2Connector(), cls.oauth2_authorization_type)()
                    except:
                        cls.access_token = getattr(OAuth2Connector(endpoint_version='v2'), cls.oauth2_authorization_type)()

    @property
    def oauth2_authorization_type(cls):
        return cls._oauth2_authorization_type

    @oauth2_authorization_type.setter
    def oauth2_authorization_type(cls, value):
        if value in ['legacy_app_flow', 'auth_code_grant', 'client_credentials_grant', 'backend_app_flow', 'web_application_flow', 'implicit_grant_flow']:
            cls._oauth2_authorization_type = value
            cls.__set_initial_property_values()

    @property
    def client_id(cls):
        return cls._client_id

    @client_id.setter
    def client_id(cls, value):
        cls._client_id = value

    @property
    def client_secret(cls):
        return cls._client_secret

    @client_secret.setter
    def client_secret(cls, value):
        cls._client_secret = value

    @property
    def tenant_id(cls):
        return cls._tenant_id

    @tenant_id.setter
    def tenant_id(cls, value):
        cls._tenant_id = value

    @property
    def access_token(cls):
        return cls._access_token

    @access_token.setter
    def access_token(cls, value):
        cls._access_token = value

    @property
    def redirect_uri(cls):
        return cls._redirect_uri

    @redirect_uri.setter
    def redirect_uri(cls, value):
        cls._redirect_uri = value

    @property
    def oauth2_scope(cls):
        return cls._oauth2_scope

    @oauth2_scope.setter
    def oauth2_scope(cls, value):
        cls._oauth2_scope = value

    @property
    def auth_header(cls):
        cls.__set_initial_property_values()
        if cls.access_token:
            cls._auth_header.update({
                'Authorization': 'Bearer {}'.format(cls.access_token)
            })
        return cls._auth_header

    @auth_header.setter
    def auth_header(cls, value):
        if value:
            cls._auth_header = {
                'X-AnchorMailbox': value
            }
        elif cls.impersonate_as != '':
            cls._auth_header = {
                'X-AnchorMailbox': cls.impersonate_as
            }
        else:
            cls._auth_header = {}

    @property
    def impersonate_as(cls):
        if not cls._impersonate_as:
            cls._impersonate_as = ''
        return cls._impersonate_as
    
    @impersonate_as.setter
    def impersonate_as(cls, value):
        if not value:
            cls._impersonate_as = ''
        else:
            cls._impersonate_as = value

    @property
    def credentials(cls):
        return cls._credentials

    @credentials.setter
    def credentials(cls, value):
        if isinstance(value, tuple):
            cls._credentials = value
            cls.__set_initial_property_values()
        else:
            raise AttributeError('Please provide both a username and password')

    @property
    def exchange_versions(cls):
        return cls._exchange_versions

    @exchange_versions.setter
    def exchange_versions(cls, value):
        if not value:
            from .exchangeversion import ExchangeVersion
            value = ExchangeVersion.EXCHANGE_VERSIONS
        if not isinstance(value, list):
            value = [value]
        cls._exchange_versions = value

    @property
    def ews_url(cls):
        return cls._ews_url

    @ews_url.setter
    def ews_url(cls, value):
        if not value:
            from .endpoints import Endpoints
            cls._ews_url = Endpoints(cls._domain).get()
        elif not isinstance(value, list):
            cls._ews_url = [value]
        else:
            cls._ews_url = value

    @property
    def domain(cls):
        return cls._domain

    @domain.setter
    def domain(cls, value):
        temp_val = None
        if '@' in value:
            local, _, domain = value.partition('@')
            temp_val = domain
        elif value:
            temp_val = value
        else:
            temp_val = None
        cls._domain = temp_val


class Authentication(object, metaclass=AuthenticationProperties):

    _auth_header = {}
    _oauth2_authorization_type = None
    _oauth2_scope = None
    _client_id = None
    _client_secret = None
    _tenant_id = None
    _access_token = None
    _impersonate_as = None
    _credentials = tuple()
    _exchange_versions = []
    _ews_url = []
    _domain = None
    _redirect_uri = 'https://google.com'



from ..service import Operation


class ExpandDL(Operation):
    """ExpandDL EWS Operation expands the provided distribution list.
    """

    RESULTS_KEY = 'Items'

    def __init__(self, user):
        """Expands the provided distribution list

        Args:
            user (str): The distribution list to expand
        """
        self.user = user

    def soap(self):
        return self.M_NAMESPACE.ExpandDL(
            self.M_NAMESPACE.Mailbox(
                self.T_NAMESPACE.EmailAddress(self.user)
            )
        )



TRAVERSAL_LIST = [
    'Deep',
    'Shallow',
    'SoftDeleted',
    'Associated'
]

FOLDER_LIST = [
        'msgfolderroot',
        'calendar',
        'contacts',
        'deleteditems',
        'drafts',
        'inbox',
        'journal',
        'notes',
        'outbox',
        'sentitems',
        'tasks',
        'junkemail',
        'searchfolders',
        'voicemail',
        'recoverableitemsdeletions',
        'recoverableitemsversions',
        'recoverableitemspurges',
        'recipientcache',
        'quickcontacts',
        'conversationhistory',
        'todosearch',
        'mycontacts',
        'imcontactlist',
        'peopleconnect',
        'favorites'
    ]

MESSAGE_ELEMENTS = {
   "MimeContent": None,
   "ItemClass": [
       'AcceptItem',
       'CalendarItem',
       'Contact',
       'Conversation',
       'DeclineItem',
       'DistributionList',
       'GlobalItemClass',
       'Item',
       'MeetingCancellation',
       'MeetingMessage',
       'MettingRequest',
       'MeetingResponse',
       'Message',
       'RemoveItem',
       'Task',
       'TentativelyAcceptItem'
   ],
   "Subject": None,
   "Sensitivity": [
       'Normal',
       'Personal',
       'Private',
       'Confidential'
   ],
   "Body": {
       'BodyType': [
           'HTML',
           'Text'
       ],
       'IsTruncated': [
           'true',
           'false'
       ]
   },
   "Attachments": {
       'FileAttachment':{
           'Name': None,
           'IsInline': ['true', 'false'],
           'Content': None
       }
   },
   "Importance": [
       'Low',
       'Normal',
       'High'
   ],
   "IsSubmitted": [
       'true',
       'false'
   ],
   "IsDraft": [
       'true',
       'false'
   ],
   "DisplayCc": None,
   "DisplayTo": None,
   "Sender": {
       'Mailbox': {
           'Name': None,
           'EmailAddress': None,
           'MailboxType': [
               'Mailbox',
               'PublicDL',
               'PrivateDl',
               'Contact',
               'PublicFolder',
               'Unknown',
               'OneOff',
               'GroupMailbox'
           ]
       }
   },
   "ToRecipients": {
       'Mailbox': {
           'Name': None,
           'EmailAddress': None,
           'MailboxType': [
               'Mailbox',
               'PublicDL',
               'PrivateDl',
               'Contact',
               'PublicFolder',
               'Unknown',
               'OneOff',
               'GroupMailbox'
           ]
       }
   },
   "CcRecipients": {
       'Mailbox': {
           'Name': None,
           'EmailAddress': None,
           'MailboxType': [
               'Mailbox',
               'PublicDL',
               'PrivateDl',
               'Contact',
               'PublicFolder',
               'Unknown',
               'OneOff',
               'GroupMailbox'
           ]
       }
   },
   "BccRecipients": {
       'Mailbox': {
           'Name': None,
           'EmailAddress': None,
           'MailboxType': [
               'Mailbox',
               'PublicDL',
               'PrivateDl',
               'Contact',
               'PublicFolder',
               'Unknown',
               'OneOff',
               'GroupMailbox'
           ]
       }
   },
   "IsReadReceiptRequested": [
       'true',
       'false'
   ],
   "IsDeliveryReceiptRequested": [
       'true',
       'false'
   ],
   "From": {
       'Mailbox': {
           'Name': None,
           'EmailAddress': None,
           'MailboxType': [
               'Mailbox',
               'PublicDL',
               'PrivateDl',
               'Contact',
               'PublicFolder',
               'Unknown',
               'OneOff',
               'GroupMailbox'
           ]
       }
   },
   "IsRead": [
       'true',
       'false'
   ],
   "IsResponseRequested": [
       'true'
   ]
}

SEARCH_FILTERS = {
    'Contains': {
        'ContainmentMode': [
            'FullString',
            'Prefixed',
            'Substring',
            'PrefixOnWords',
            'ExactPhrase'
        ],
        'ContainmentComparison': [
            'Exact',
            'IgnoreCase',
            'IgnoreNonSpacingCharacters',
            'Loose',
            'IgnoreCaseAndNonSpacingCharacters',
            'LooseAndIgnoreCase',
            'LooseAndIgnoreNonSpace',
            'LooseAndIgnoreCaseAndIgnoreNonSpace'
        ]
    },
    'Excludes': {
        'Bitmask': []
    },
    'Exists': {},
    'IsEqualTo': {
        'FieldURIOrConstant': []
    },
    'IsNotEqualTo': {},
    'IsGreaterThan': {},
    'IsGreaterThanOrEqualTo': {},
    'IsLessThan': {},
    'IsLessThanOrEqualTo': {}
}

FIELD_URI_MAP = {
    'folder': [
        "FolderId",
        "ParentFolderId",
        "DisplayName",
        "UnreadCount",
        "TotalCount",
        "ChildFolderCount",
        "FolderClass",
        "SearchParameters",
        "ManagedFolderInformation",
        "PermissionSet",
        "EffectiveRights",
        "SharingEffectiveRights"
    ],
    'item': [
        "ItemId",
        "ParentFolderId",
        "ItemClass",
        "MimeContent",
        "Attachments",
        "Subject",
        "DateTimeReceived",
        "Size",
        "Categories",
        "HasAttachments",
        "Importance",
        "InReplyTo",
        "InternetMessageHeaders",
        "IsAssociated",
        "IsDraft",
        "IsFromMe",
        "IsResend",
        "IsSubmitted",
        "IsUnmodified",
        "DateTimeSent",
        "DateTimeCreated",
        "Body",
        "ResponseObjects",
        "Sensitivity",
        "ReminderDueBy",
        "ReminderIsSet",
        "ReminderNextTime",
        "ReminderMinutesBeforeStart",
        "DisplayTo",
        "DisplayCc",
        "Culture",
        "EffectiveRights",
        "LastModifiedName",
        "LastModifiedTime",
        "ConversationId",
        "UniqueBody",
        "Flag",
        "StoreEntryId",
        "InstanceKey",
        "NormalizedBody",
        "EntityExtractionResult",
        "ArchiveTag",
        "RetentionDate",
        "Preview",
        "NextPredictedAction",
        "GroupingAction",
        "PredictedActionReasons",
        "IsClutter",
        "RightsManagementLicenseData",
        "BlockStatus",
        "HasBlockedImages",
        "WebClientReadFormQueryString",
        "WebClientEditFormQueryString",
        "TextBody",
        "IconIndex",
        "MimeContentUTF8"
    ],
    'message': [
        "ConversationIndex",
        "ConversationTopic",
        "InternetMessageId",
        "IsRead",
        "IsResponseRequested",
        "IsReadReceiptRequested",
        "IsDeliveryReceiptRequested",
        "ReceivedBy",
        "ReceivedRepresenting",
        "References",
        "ReplyTo",
        "From",
        "Sender",
        "ToRecipients",
        "CcRecipients",
        "BccRecipients",
        "ApprovalRequestData",
        "VotingInformation",
        "ReminderMessageData",
    ],
    'meeting': [
        "AssociatedCalendarItemId",
        "IsDelegated",
        "IsOutOfDate",
        "HasBeenProcessed",
        "ResponseType",
        "ProposedStart",
        "PropsedEnd",
    ],
    'meetingRequest': [
        "MeetingRequestType",
        "IntendedFreeBusyStatus",
        "ChangeHighlights",
    ],
    'calendar': [
        "Start",
        "End",
        "OriginalStart",
        "StartWallClock",
        "EndWallClock",
        "StartTimeZoneId",
        "EndTimeZoneId",
        "IsAllDayEvent",
        "LegacyFreeBusyStatus",
        "Location",
        "When",
        "IsMeeting",
        "IsCancelled",
        "IsRecurring",
        "MeetingRequestWasSent",
        "IsResponseRequested",
        "CalendarItemType",
        "MyResponseType",
        "Organizer",
        "RequiredAttendees",
        "OptionalAttendees",
        "Resources",
        "ConflictingMeetingCount",
        "AdjacentMeetingCount",
        "ConflictingMeetings",
        "AdjacentMeetings",
        "Duration",
        "TimeZone",
        "AppointmentReplyTime",
        "AppointmentSequenceNumber",
        "AppointmentState",
        "Recurrence",
        "FirstOccurrence",
        "LastOccurrence",
        "ModifiedOccurrences",
        "DeletedOccurrences",
        "MeetingTimeZone",
        "ConferenceType",
        "AllowNewTimeProposal",
        "IsOnlineMeeting",
        "MeetingWorkspaceUrl",
        "NetShowUrl",
        "UID",
        "RecurrenceId",
        "DateTimeStamp",
        "StartTimeZone",
        "EndTimeZone",
        "JoinOnlineMeetingUrl",
        "OnlineMeetingSettings",
        "IsOrganizer",
    ],
    'task': [
        "ActualWork",
        "AssignedTime",
        "BillingInformation",
        "ChangeCount",
        "Companies",
        "CompleteDate",
        "Contacts",
        "DelegationState",
        "Delegator",
        "DueDate",
        "IsAssignmentEditable",
        "IsComplete",
        "IsRecurring",
        "IsTeamTask",
        "Mileage",
        "Owner",
        "PercentComplete",
        "Recurrence",
        "StartDate",
        "Status",
        "StatusDescription",
        "TotalWork",
    ],
    'contacts': [
        "Alias",
        "AssistantName",
        "Birthday",
        "BusinessHomePage",
        "Children",
        "Companies",
        "CompanyName",
        "CompleteName",
        "ContactSource",
        "Culture",
        "Department",
        "DisplayName",
        "DirectoryId",
        "DirectReports",
        "EmailAddresses",
        "FileAs",
        "FileAsMapping",
        "Generation",
        "GivenName",
        "ImAddresses",
        "Initials",
        "JobTitle",
        "Manager",
        "ManagerMailbox",
        "MiddleName",
        "Mileage",
        "MSExchangeCertificate",
        "Nickname",
        "Notes",
        "OfficeLocation",
        "PhoneNumbers",
        "PhoneticFullName",
        "PhoneticFirstName",
        "PhoneticLastName",
        "Photo",
        "PhysicalAddresses",
        "PostalAddressIndex",
        "Profession",
        "SpouseName",
        "Surname",
        "WeddingAnniversary",
        "UserSMIMECertificate",
        "HasPicture",
    ],
    'conversation': [
        "ConversationId",
        "ConversationTopic",
        "UniqueRecipients",
        "GlobalUniqueRecipients",
        "UniqueUnreadSenders",
        "GlobalUniqueUnreadSenders",
        "UniqueSenders",
        "GlobalUniqueSenders",
        "LastDeliveryTime",
        "GlobalLastDeliveryTime",
        "Categories",
        "GlobalCategories",
        "FlagStatus",
        "GlobalFlagStatus",
        "HasAttachments",
        "GlobalHasAttachments",
        "HasIrm",
        "GlobalHasIrm",
        "MessageCount",
        "GlobalMessageCount",
        "UnreadCount",
        "GlobalUnreadCount",
        "Size",
        "GlobalSize",
        "ItemClasses",
        "GlobalItemClasses",
        "Importance",
        "GlobalImportance",
        "ItemIds",
        "GlobalItemIds",
        "LastModifiedTime",
        "InstanceKey",
        "Preview",
        "GlobalParentFolderId",
        "NextPredictedAction",
        "GroupingAction",
        "IconIndex",
        "GlobalIconIndex",
        "DraftItemIds",
        "HasClutter",
    ],
    'persona': [
        "PersonaId",
        "PersonaType",
        "GivenName",
        "CompanyName",
        "Surname",
        "DisplayName",
        "EmailAddress",
        "FileAs",
        "HomeCity",
        "CreationTime",
        "RelevanceScore",
        "WorkCity",
        "PersonaObjectStatus",
        "FileAsId",
        "DisplayNamePrefix",
        "YomiCompanyName",
        "YomiFirstName",
        "YomiLastName",
        "Title",
        "EmailAddresses",
        "PhoneNumber",
        "ImAddress",
        "ImAddresses",
        "ImAddresses2",
        "ImAddresses3",
        "FolderIds",
        "Attributions",
        "DisplayNames",
        "Initials",
        "FileAses",
        "FileAsIds",
        "DisplayNamePrefixes",
        "GivenNames",
        "MiddleNames",
        "Surnames",
        "Generations",
        "Nicknames",
        "YomiCompanyNames",
        "YomiFirstNames",
        "YomiLastNames",
        "BusinessPhoneNumbers",
        "BusinessPhoneNumbers2",
        "HomePhones",
        "HomePhones2",
        "MobilePhones",
        "MobilePhones2",
        "AssistantPhoneNumbers",
        "CallbackPhones",
        "CarPhones",
        "HomeFaxes",
        "OrganizationMainPhones",
        "OtherFaxes",
        "OtherTelephones",
        "OtherPhones2",
        "Pagers",
        "RadioPhones",
        "TelexNumbers",
        "WorkFaxes",
        "Emails1",
        "Emails2",
        "Emails3",
        "BusinessHomePages",
        "School",
        "PersonalHomePages",
        "OfficeLocations",
        "BusinessAddresses",
        "HomeAddresses",
        "OtherAddresses",
        "Titles",
        "Departments",
        "CompanyNames",
        "Managers",
        "AssistantNames",
        "Professions",
        "SpouseNames",
        "Hobbies",
        "WeddingAnniversaries",
        "Birthdays",
        "Children",
        "Locations",
        "ExtendedProperties",
        "PostalAddress",
        "Bodies",
    ]
}

RESPONSE_CODES = [
    "NoError",
    "ErrorAccessDenied",
    "ErrorAccessModeSpecified",
    "ErrorAccountDisabled",
    "ErrorAddDelegatesFailed",
    "ErrorAddressSpaceNotFound",
    "ErrorADOperation",
    "ErrorADSessionFilter",
    "ErrorADUnavailable",
    "ErrorServiceUnavailable",
    "ErrorAutoDiscoverFailed",
    "ErrorAffectedTaskOccurrencesRequired",
    "ErrorAttachmentNestLevelLimitExceeded" ,
    "ErrorAttachmentSizeLimitExceeded",
    "ErrorArchiveFolderPathCreation",
    "ErrorArchiveMailboxNotEnabled",
    "ErrorArchiveMailboxServiceDiscoveryFailed",
    "ErrorAvailabilityConfigNotFound",
    "ErrorBatchProcessingStopped",
    "ErrorCalendarCannotMoveOrCopyOccurrence",
    "ErrorCalendarCannotUpdateDeletedItem",
    "ErrorCalendarCannotUseIdForOccurrenceId",
    "ErrorCalendarCannotUseIdForRecurringMasterId",
    "ErrorCalendarDurationIsTooLong",
    "ErrorCalendarEndDateIsEarlierThanStartDate",
    "ErrorCalendarFolderIsInvalidForCalendarView",
    "ErrorCalendarInvalidAttributeValue",
    "ErrorCalendarInvalidDayForTimeChangePattern",
    "ErrorCalendarInvalidDayForWeeklyRecurrence",
    "ErrorCalendarInvalidPropertyState",
    "ErrorCalendarInvalidPropertyValue",
    "ErrorCalendarInvalidRecurrence",
    "ErrorCalendarInvalidTimeZone",
    "ErrorCalendarIsCancelledForAccept",
    "ErrorCalendarIsCancelledForDecline",
    "ErrorCalendarIsCancelledForRemove",
    "ErrorCalendarIsCancelledForTentative",
    "ErrorCalendarIsDelegatedForAccept",
    "ErrorCalendarIsDelegatedForDecline",
    "ErrorCalendarIsDelegatedForRemove",
    "ErrorCalendarIsDelegatedForTentative",
    "ErrorCalendarIsNotOrganizer",
    "ErrorCalendarIsOrganizerForAccept",
    "ErrorCalendarIsOrganizerForDecline",
    "ErrorCalendarIsOrganizerForRemove",
    "ErrorCalendarIsOrganizerForTentative",
    "ErrorCalendarOccurrenceIndexIsOutOfRecurrenceRange",
    "ErrorCalendarOccurrenceIsDeletedFromRecurrence",
    "ErrorCalendarOutOfRange",
    "ErrorCalendarMeetingRequestIsOutOfDate",
    "ErrorCalendarViewRangeTooBig",
    "ErrorCallerIsInvalidADAccount",
    "ErrorCannotAccessDeletedPublicFolder",
    "ErrorCannotArchiveCalendarContactTaskFolderException",
    "ErrorCannotArchiveItemsInPublicFolders",
    "ErrorCannotArchiveItemsInArchiveMailbox",
    "ErrorCannotCreateCalendarItemInNonCalendarFolder",
    "ErrorCannotCreateContactInNonContactFolder",
    "ErrorCannotCreatePostItemInNonMailFolder",
    "ErrorCannotCreateTaskInNonTaskFolder",
    "ErrorCannotDeleteObject",
    "ErrorCannotDisableMandatoryExtension",
    "ErrorCannotFindUser",
    "ErrorCannotGetSourceFolderPath",
    "ErrorCannotGetExternalEcpUrl",
    "ErrorCannotOpenFileAttachment",
    "ErrorCannotDeleteTaskOccurrence",
    "ErrorCannotEmptyFolder",
    "ErrorCannotSetCalendarPermissionOnNonCalendarFolder",
    "ErrorCannotSetNonCalendarPermissionOnCalendarFolder",
    "ErrorCannotSetPermissionUnknownEntries",
    "ErrorCannotSpecifySearchFolderAsSourceFolder",
    "ErrorCannotUseFolderIdForItemId",
    "ErrorCannotUseItemIdForFolderId",
    "ErrorChangeKeyRequired",
    "ErrorChangeKeyRequiredForWriteOperations",
    "ErrorClientDisconnected",
    "ErrorClientIntentInvalidStateDefinition",
    "ErrorClientIntentNotFound",
    "ErrorConnectionFailed",
    "ErrorContainsFilterWrongType",
    "ErrorContentConversionFailed",
    "ErrorContentIndexingNotEnabled",
    "ErrorCorruptData",
    "ErrorCreateItemAccessDenied",
    "ErrorCreateManagedFolderPartialCompletion",
    "ErrorCreateSubfolderAccessDenied",
    "ErrorCrossMailboxMoveCopy",
    "ErrorCrossSiteRequest",
    "ErrorDataSizeLimitExceeded",
    "ErrorDataSourceOperation",
    "ErrorDelegateAlreadyExists",
    "ErrorDelegateCannotAddOwner",
    "ErrorDelegateMissingConfiguration",
    "ErrorDelegateNoUser",
    "ErrorDelegateValidationFailed",
    "ErrorDeleteDistinguishedFolder",
    "ErrorDeleteItemsFailed",
    "ErrorDeleteUnifiedMessagingPromptFailed",
    "ErrorDistinguishedUserNotSupported",
    "ErrorDistributionListMemberNotExist",
    "ErrorDuplicateInputFolderNames",
    "ErrorDuplicateUserIdsSpecified",
    "ErrorDuplicateTransactionId",
    "ErrorEmailAddressMismatch",
    "ErrorEventNotFound",
    "ErrorExceededConnectionCount",
    "ErrorExceededSubscriptionCount",
    "ErrorExceededFindCountLimit",
    "ErrorExpiredSubscription",
    "ErrorExtensionNotFound",
    "ErrorExtensionsNotAuthorized",
    "ErrorFolderCorrupt",
    "ErrorFolderNotFound",
    "ErrorFolderPropertRequestFailed",
    "ErrorFolderSave",
    "ErrorFolderSaveFailed",
    "ErrorFolderSavePropertyError",
    "ErrorFolderExists",
    "ErrorFreeBusyGenerationFailed",
    "ErrorGetServerSecurityDescriptorFailed",
    "ErrorImContactLimitReached",
    "ErrorImGroupDisplayNameAlreadyExists",
    "ErrorImGroupLimitReached",
    "ErrorImpersonateUserDenied",
    "ErrorImpersonationDenied",
    "ErrorImpersonationFailed",
    "ErrorIncorrectSchemaVersion",
    "ErrorIncorrectUpdatePropertyCount",
    "ErrorIndividualMailboxLimitReached",
    "ErrorInsufficientResources",
    "ErrorInternalServerError",
    "ErrorInternalServerTransientError",
    "ErrorInvalidAccessLevel",
    "ErrorInvalidArgument",
    "ErrorInvalidAttachmentId",
    "ErrorInvalidAttachmentSubfilter",
    "ErrorInvalidAttachmentSubfilterTextFilter",
    "ErrorInvalidAuthorizationContext",
    "ErrorInvalidChangeKey",
    "ErrorInvalidClientSecurityContext",
    "ErrorInvalidCompleteDate",
    "ErrorInvalidContactEmailAddress",
    "ErrorInvalidContactEmailIndex",
    "ErrorInvalidCrossForestCredentials",
    "ErrorInvalidDelegatePermission",
    "ErrorInvalidDelegateUserId",
    "ErrorInvalidExcludesRestriction",
    "ErrorInvalidExpressionTypeForSubFilter",
    "ErrorInvalidExtendedProperty",
    "ErrorInvalidExtendedPropertyValue",
    "ErrorInvalidFolderId",
    "ErrorInvalidFolderTypeForOperation",
    "ErrorInvalidFractionalPagingParameters",
    "ErrorInvalidFreeBusyViewType",
    "ErrorInvalidId",
    "ErrorInvalidIdEmpty",
    "ErrorInvalidIdMalformed",
    "ErrorInvalidIdMalformedEwsLegacyIdFormat",
    "ErrorInvalidIdMonikerTooLong",
    "ErrorInvalidIdNotAnItemAttachmentId",
    "ErrorInvalidIdReturnedByResolveNames",
    "ErrorInvalidIdStoreObjectIdTooLong",
    "ErrorInvalidIdTooManyAttachmentLevels",
    "ErrorInvalidIdXml",
    "ErrorInvalidImContactId",
    "ErrorInvalidImDistributionGroupSmtpAddress",
    "ErrorInvalidImGroupId",
    "ErrorInvalidIndexedPagingParameters",
    "ErrorInvalidInternetHeaderChildNodes",
    "ErrorInvalidItemForOperationArchiveItem",
    "ErrorInvalidItemForOperationCreateItemAttachment",
    "ErrorInvalidItemForOperationCreateItem",
    "ErrorInvalidItemForOperationAcceptItem",
    "ErrorInvalidItemForOperationDeclineItem",
    "ErrorInvalidItemForOperationCancelItem",
    "ErrorInvalidItemForOperationExpandDL",
    "ErrorInvalidItemForOperationRemoveItem",
    "ErrorInvalidItemForOperationSendItem",
    "ErrorInvalidItemForOperationTentative",
    "ErrorInvalidLogonType",
    "ErrorInvalidLikeRequest",
    "ErrorInvalidMailbox",
    "ErrorInvalidManagedFolderProperty",
    "ErrorInvalidManagedFolderQuota",
    "ErrorInvalidManagedFolderSize",
    "ErrorInvalidMergedFreeBusyInterval",
    "ErrorInvalidNameForNameResolution",
    "ErrorInvalidOperation",
    "ErrorInvalidNetworkServiceContext",
    "ErrorInvalidOofParameter",
    "ErrorInvalidPagingMaxRows",
    "ErrorInvalidParentFolder",
    "ErrorInvalidPercentCompleteValue",
    "ErrorInvalidPermissionSettings",
    "ErrorInvalidPhoneCallId",
    "ErrorInvalidPhoneNumber",
    "ErrorInvalidUserInfo",
    "ErrorInvalidPropertyAppend",
    "ErrorInvalidPropertyDelete",
    "ErrorInvalidPropertyForExists",
    "ErrorInvalidPropertyForOperation",
    "ErrorInvalidPropertyRequest",
    "ErrorInvalidPropertySet",
    "ErrorInvalidPropertyUpdateSentMessage",
    "ErrorInvalidProxySecurityContext",
    "ErrorInvalidPullSubscriptionId",
    "ErrorInvalidPushSubscriptionUrl",
    "ErrorInvalidRecipients",
    "ErrorInvalidRecipientSubfilter",
    "ErrorInvalidRecipientSubfilterComparison",
    "ErrorInvalidRecipientSubfilterOrder",
    "ErrorInvalidRecipientSubfilterTextFilter",
    "ErrorInvalidReferenceItem",
    "ErrorInvalidRequest",
    "ErrorInvalidRestriction",
    "ErrorInvalidRetentionTagTypeMismatch",
    "ErrorInvalidRetentionTagInvisible",
    "ErrorInvalidRetentionTagInheritance",
    "ErrorInvalidRetentionTagIdGuid",
    "ErrorInvalidRoutingType",
    "ErrorInvalidScheduledOofDuration",
    "ErrorInvalidSchemaVersionForMailboxVersion",
    "ErrorInvalidSecurityDescriptor",
    "ErrorInvalidSendItemSaveSettings",
    "ErrorInvalidSerializedAccessToken",
    "ErrorInvalidServerVersion",
    "ErrorInvalidSid",
    "ErrorInvalidSIPUri",
    "ErrorInvalidSmtpAddress",
    "ErrorInvalidSubfilterType",
    "ErrorInvalidSubfilterTypeNotAttendeeType",
    "ErrorInvalidSubfilterTypeNotRecipientType",
    "ErrorInvalidSubscription",
    "ErrorInvalidSubscriptionRequest",
    "ErrorInvalidSyncStateData",
    "ErrorInvalidTimeInterval",
    "ErrorInvalidUserOofSettings",
    "ErrorInvalidUserPrincipalName",
    "ErrorInvalidUserSid",
    "ErrorInvalidUserSidMissingUPN",
    "ErrorInvalidValueForProperty",
    "ErrorInvalidWatermark",
    "ErrorIPGatewayNotFound",
    "ErrorIrresolvableConflict",
    "ErrorItemCorrupt",
    "ErrorItemNotFound",
    "ErrorItemPropertyRequestFailed",
    "ErrorItemSave",
    "ErrorItemSavePropertyError",
    "ErrorLegacyMailboxFreeBusyViewTypeNotMerged",
    "ErrorLocalServerObjectNotFound",
    "ErrorLogonAsNetworkServiceFailed",
    "ErrorMailboxConfiguration",
    "ErrorMailboxDataArrayEmpty",
    "ErrorMailboxDataArrayTooBig",
    "ErrorMailboxHoldNotFound",
    "ErrorMailboxLogonFailed",
    "ErrorMailboxMoveInProgress",
    "ErrorMailboxStoreUnavailable",
    "ErrorMailRecipientNotFound",
    "ErrorMailTipsDisabled",
    "ErrorManagedFolderAlreadyExists",
    "ErrorManagedFolderNotFound",
    "ErrorManagedFoldersRootFailure",
    "ErrorMeetingSuggestionGenerationFailed",
    "ErrorMessageDispositionRequired",
    "ErrorMessageSizeExceeded",
    "ErrorMimeContentConversionFailed",
    "ErrorMimeContentInvalid",
    "ErrorMimeContentInvalidBase64String",
    "ErrorMissingArgument",
    "ErrorMissingEmailAddress",
    "ErrorMissingEmailAddressForManagedFolder",
    "ErrorMissingInformationEmailAddress",
    "ErrorMissingInformationReferenceItemId",
    "ErrorMissingItemForCreateItemAttachment",
    "ErrorMissingManagedFolderId",
    "ErrorMissingRecipients",
    "ErrorMissingUserIdInformation",
    "ErrorMoreThanOneAccessModeSpecified",
    "ErrorMoveCopyFailed",
    "ErrorMoveDistinguishedFolder",
    "ErrorMoveUnifiedGroupPropertyFailed",
    "ErrorMultiLegacyMailboxAccess",
    "ErrorNameResolutionMultipleResults",
    "ErrorNameResolutionNoMailbox",
    "ErrorNameResolutionNoResults",
    "ErrorNoApplicableProxyCASServersAvailable",
    "ErrorNoCalendar",
    "ErrorNoDestinationCASDueToKerberosRequirements",
    "ErrorNoDestinationCASDueToSSLRequirements",
    "ErrorNoDestinationCASDueToVersionMismatch",
    "ErrorNoFolderClassOverride",
    "ErrorNoFreeBusyAccess",
    "ErrorNonExistentMailbox",
    "ErrorNonPrimarySmtpAddress",
    "ErrorNoPropertyTagForCustomProperties",
    "ErrorNoPublicFolderReplicaAvailable",
    "ErrorNoPublicFolderServerAvailable",
    "ErrorNoRespondingCASInDestinationSite",
    "ErrorNotDelegate",
    "ErrorNotEnoughMemory",
    "ErrorObjectTypeChanged",
    "ErrorOccurrenceCrossingBoundary",
    "ErrorOccurrenceTimeSpanTooBig" ,
    "ErrorOperationNotAllowedWithPublicFolderRoot" ,
    "ErrorParentFolderIdRequired",
    "ErrorParentFolderNotFound",
    "ErrorPasswordChangeRequired",
    "ErrorPasswordExpired",
    "ErrorPhoneNumberNotDialable",
    "ErrorPropertyUpdate",
    "ErrorPromptPublishingOperationFailed",
    "ErrorPropertyValidationFailure",
    "ErrorProxiedSubscriptionCallFailure",
    "ErrorProxyCallFailed",
    "ErrorProxyGroupSidLimitExceeded",
    "ErrorProxyRequestNotAllowed",
    "ErrorProxyRequestProcessingFailed",
    "ErrorProxyServiceDiscoveryFailed",
    "ErrorProxyTokenExpired",
    "ErrorPublicFolderMailboxDiscoveryFailed",
    "ErrorPublicFolderOperationFailed",
    "ErrorPublicFolderRequestProcessingFailed",
    "ErrorPublicFolderServerNotFound",
    "ErrorPublicFolderSyncException",
    "ErrorQueryFilterTooLong",
    "ErrorQuotaExceeded",
    "ErrorReadEventsFailed",
    "ErrorReadReceiptNotPending",
    "ErrorRecurrenceEndDateTooBig",
    "ErrorRecurrenceHasNoOccurrence",
    "ErrorRemoveDelegatesFailed",
    "ErrorRequestAborted",
    "ErrorRequestStreamTooBig",
    "ErrorRequiredPropertyMissing",
    "ErrorResolveNamesInvalidFolderType",
    "ErrorResolveNamesOnlyOneContactsFolderAllowed",
    "ErrorResponseSchemaValidation",
    "ErrorRestrictionTooLong",
    "ErrorRestrictionTooComplex",
    "ErrorResultSetTooBig",
    "ErrorInvalidExchangeImpersonationHeaderData",
    "ErrorSavedItemFolderNotFound",
    "ErrorSchemaValidation",
    "ErrorSearchFolderNotInitialized",
    "ErrorSendAsDenied",
    "ErrorSendMeetingCancellationsRequired",
    "ErrorSendMeetingInvitationsOrCancellationsRequired",
    "ErrorSendMeetingInvitationsRequired",
    "ErrorSentMeetingRequestUpdate",
    "ErrorSentTaskRequestUpdate",
    "ErrorServerBusy",
    "ErrorServiceDiscoveryFailed",
    "ErrorStaleObject",
    "ErrorSubmissionQuotaExceeded",
    "ErrorSubscriptionAccessDenied",
    "ErrorSubscriptionDelegateAccessNotSupported",
    "ErrorSubscriptionNotFound",
    "ErrorSubscriptionUnsubscribed",
    "ErrorSyncFolderNotFound",
    "ErrorTeamMailboxNotFound",
    "ErrorTeamMailboxNotLinkedToSharePoint",
    "ErrorTeamMailboxUrlValidationFailed",
    "ErrorTeamMailboxNotAuthorizedOwner",
    "ErrorTeamMailboxActiveToPendingDelete",
    "ErrorTeamMailboxFailedSendingNotifications",
    "ErrorTeamMailboxErrorUnknown",
    "ErrorTimeIntervalTooBig",
    "ErrorTimeoutExpired",
    "ErrorTimeZone",
    "ErrorToFolderNotFound",
    "ErrorTokenSerializationDenied",
    "ErrorTooManyObjectsOpened",
    "ErrorUpdatePropertyMismatch",
    "ErrorAccessingPartialCreatedUnifiedGroup",
    "ErrorUnifiedGroupMailboxAADCreationFailed",
    "ErrorUnifiedGroupMailboxAADDeleteFailed",
    "ErrorUnifiedGroupMailboxNamingPolicy",
    "ErrorUnifiedGroupMailboxDeleteFailed",
    "ErrorUnifiedGroupMailboxNotFound",
    "ErrorUnifiedGroupMailboxUpdateDelayed",
    "ErrorUnifiedGroupMailboxUpdatedPartialProperties",
    "ErrorUnifiedGroupMailboxUpdateFailed",
    "ErrorUnifiedGroupMailboxProvisionFailed",
    "ErrorUnifiedMessagingDialPlanNotFound",
    "ErrorUnifiedMessagingReportDataNotFound",
    "ErrorUnifiedMessagingPromptNotFound",
    "ErrorUnifiedMessagingRequestFailed",
    "ErrorUnifiedMessagingServerNotFound",
    "ErrorUnableToGetUserOofSettings",
    "ErrorUnableToRemoveImContactFromGroup",
    "ErrorUnsupportedSubFilter",
    "ErrorUnsupportedCulture",
    "ErrorUnsupportedMapiPropertyType",
    "ErrorUnsupportedMimeConversion",
    "ErrorUnsupportedPathForQuery",
    "ErrorUnsupportedPathForSortGroup",
    "ErrorUnsupportedPropertyDefinition",
    "ErrorUnsupportedQueryFilter",
    "ErrorUnsupportedRecurrence",
    "ErrorUnsupportedTypeForConversion",
    "ErrorUpdateDelegatesFailed",
    "ErrorUserNotUnifiedMessagingEnabled",
    "ErrorVoiceMailNotImplemented",
    "ErrorValueOutOfRange",
    "ErrorVirusDetected",
    "ErrorVirusMessageDeleted",
    "ErrorWebRequestInInvalidState",
    "ErrorWin32InteropError",
    "ErrorWorkingHoursSaveFailed",
    "ErrorWorkingHoursXmlMalformed",
    "ErrorWrongServerVersion",
    "ErrorWrongServerVersionDelegate",
    "ErrorMissingInformationSharingFolderId",
    "ErrorDuplicateSOAPHeader" ,
    "ErrorSharingSynchronizationFailed" ,
    "ErrorSharingNoExternalEwsAvailable" ,
    "ErrorFreeBusyDLLimitReached",
    "ErrorInvalidGetSharingFolderRequest" ,
    "ErrorNotAllowedExternalSharingByPolicy" ,
    "ErrorUserNotAllowedByPolicy" ,
    "ErrorPermissionNotAllowedByPolicy" ,
    "ErrorOrganizationNotFederated" ,
    "ErrorMailboxFailover" ,
    "ErrorInvalidExternalSharingInitiator" ,
    "ErrorMessageTrackingPermanentError" ,
    "ErrorMessageTrackingTransientError" ,
    "ErrorMessageTrackingNoSuchDomain" ,
    "ErrorUserWithoutFederatedProxyAddress" ,
    "ErrorInvalidOrganizationRelationshipForFreeBusy" ,
    "ErrorInvalidFederatedOrganizationId" ,
    "ErrorInvalidExternalSharingSubscriber" ,
    "ErrorInvalidSharingData" ,
    "ErrorInvalidSharingMessage" ,
    "ErrorNotSupportedSharingMessage" ,
    "ErrorApplyConversationActionFailed" ,
    "ErrorInboxRulesValidationError" ,
    "ErrorOutlookRuleBlobExists" ,
    "ErrorRulesOverQuota" ,
    "ErrorNewEventStreamConnectionOpened" ,
    "ErrorMissedNotificationEvents" ,
    "ErrorDuplicateLegacyDistinguishedName" ,
    "ErrorInvalidClientAccessTokenRequest" ,
    "ErrorUnauthorizedClientAccessTokenRequest" ,
    "ErrorNoSpeechDetected" ,
    "ErrorUMServerUnavailable" ,
    "ErrorRecipientNotFound" ,
    "ErrorRecognizerNotInstalled" ,
    "ErrorSpeechGrammarError" ,
    "ErrorInvalidManagementRoleHeader" ,
    "ErrorLocationServicesDisabled",
    "ErrorLocationServicesRequestTimedOut",
    "ErrorLocationServicesRequestFailed",
    "ErrorLocationServicesInvalidRequest",
    "ErrorWeatherServiceDisabled",
    "ErrorMailboxScopeNotAllowedWithoutQueryString" ,
    "ErrorArchiveMailboxSearchFailed" ,
    "ErrorGetRemoteArchiveFolderFailed" ,
    "ErrorFindRemoteArchiveFolderFailed" ,
    "ErrorGetRemoteArchiveItemFailed" ,
    "ErrorExportRemoteArchiveItemsFailed" ,
    "ErrorInvalidPhotoSize" ,
    "ErrorSearchQueryHasTooManyKeywords",
    "ErrorSearchTooManyMailboxes",
    "ErrorInvalidRetentionTagNone",
    "ErrorDiscoverySearchesDisabled",
    "ErrorCalendarSeekToConditionNotSupported",
    "ErrorCalendarIsGroupMailboxForAccept",
    "ErrorCalendarIsGroupMailboxForDecline",
    "ErrorCalendarIsGroupMailboxForTentative",
    "ErrorCalendarIsGroupMailboxForSuppressReadReceipt",
    "ErrorOrganizationAccessBlocked",
    "ErrorInvalidLicense",
    "ErrorMessagePerFolderCountReceiveQuotaExceeded",
    "ErrorInvalidBulkActionType",
    "ErrorInvalidKeepNCount",
    "ErrorInvalidKeepNType",
    "ErrorNoOAuthServerAvailableForRequest",  
    "ErrorInstantSearchSessionExpired",
    "ErrorInstantSearchTimeout",
    "ErrorInstantSearchFailed",
    "ErrorUnsupportedUserForExecuteSearch",
    "ErrorDuplicateExtendedKeywordDefinition",
    "ErrorMissingExchangePrincipal",
    "ErrorUnexpectedUnifiedGroupsCount",
    "ErrorParsingXMLResponse",
    "ErrorInvalidFederationOrganizationIdentifier",
    "ErrorInvalidSweepRule",
    "ErrorInvalidSweepRuleOperationType",
    "ErrorTargetDomainNotSupported",
    "ErrorInvalidInternetWebProxyOnLocalServer",
    "ErrorNoSenderRestrictionsSettingsFoundInRequest",
    "ErrorDuplicateSenderRestrictionsInputFound",
    "ErrorSenderRestrictionsUpdateFailed",
    "ErrorMessageSubmissionBlocked",
    "ErrorExceededMessageLimit",
    "ErrorExceededMaxRecipientLimitBlock",
    "ErrorAccountSuspend",
    "ErrorExceededMaxRecipientLimit",
    "ErrorMessageBlocked",
    "ErrorAccountSuspendShowTierUpgrade",
    "ErrorExceededMessageLimitShowTierUpgrade",
    "ErrorExceededMaxRecipientLimitShowTierUpgrade",
    "ErrorInvalidLongitude",
    "ErrorInvalidLatitude",
    "ErrorProxySoapException",
    "ErrorUnifiedGroupAlreadyExists",
    "ErrorUnifiedGroupAadAuthorizationRequestDenied",
    "ErrorUnifiedGroupCreationDisabled",
    "ErrorMarketPlaceExtensionAlreadyInstalledForOrg",
    "ErrorExtensionAlreadyInstalledForOrg",
    "ErrorNewerExtensionAlreadyInstalled",
    "ErrorNewerMarketPlaceExtensionAlreadyInstalled",
    "ErrorInvalidExtensionId",
    "ErrorCannotUninstallProvidedExtensions",
    "ErrorNoRbacPermissionToInstallMarketPlaceExtensions",
    "ErrorNoRbacPermissionToInstallReadWriteMailboxExtensions",
    "ErrorInvalidReportMessageActionType",
    "ErrorCannotDownloadExtensionManifest",
    "ErrorCalendarForwardActionNotAllowed",
    "ErrorUnifiedGroupAliasNamingPolicy",
    "ErrorSubscriptionsDisabledForGroup",
    "ErrorCannotFindFileAttachment",
    "ErrorInvalidValueForFilter",
    "ErrorQuotaExceededOnDelete",
    "ErrorAccessDeniedDueToCompliance",
    "ErrorRecoverableItemsAccessDenied"
]


import os
import collections

from ..core import Core
from ..endpoint import SyncFolderItems, SyncFolderHierarchy, GetItem, ConvertId


class HierarchyProcessor:

    __FOLDER_HIERARCHY = []

    def __get_item(self, item_id, change_key=None):
        response = GetItem(item_id, change_key=change_key).run()
        if isinstance(response, list):
            if any(item in response for item in ConvertId.ID_FORMATS):
                convert_id_response = ConvertId(Core.credentials[0], item_id, id_type=response[0], convert_to=response[1]).run()
                get_item_response = GetItem(convert_id_response[0]).run()
                return get_item_response if get_item_response else None
        return GetItem(item_id, change_key=change_key).run()

    def __dict_merge(self, dct, merge_dct, add_keys=True):
        """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
        updating only top-level keys, dict_merge recurses down into dicts nested
        to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
        ``dct``.

        This version will return a copy of the dictionary and leave the original
        arguments untouched.

        The optional argument ``add_keys``, determines whether keys which are
        present in ``merge_dict`` but not ``dct`` should be included in the
        new dict.

        Args:
            dct (dict) onto which the merge is executed
            merge_dct (dict): dct merged into dct
            add_keys (bool): whether to add new keys

        Returns:
            dict: updated dict
        """
        dct = dct.copy()
        if not add_keys:
            merge_dct = {
                k: merge_dct[k]
                for k in set(dct).intersection(set(merge_dct))
            }
        for k, v in merge_dct.items():
            if (k in dct and isinstance(dct[k], dict)
                    and isinstance(merge_dct[k], collections.Mapping)):
                dct[k] = self.__dict_merge(dct[k], merge_dct[k], add_keys=add_keys)
            else:
                dct[k] = merge_dct[k]
        return dct

    def __get_children(self, parent, relations):
        children = (r[1] for r in relations if r[0] == parent)
        return {c: self.__get_children(c, relations) for c in children}

    def __sync_folder_items(self, tree):
        if isinstance(tree, dict):
            for key, val in tree.items():
                if key not in self.__FOLDER_HIERARCHY:
                    self.__FOLDER_HIERARCHY.append(key)
                if isinstance(val, dict):
                    self.__sync_folder_items(val)

    def __extract_item_id(self, data):
        for key,val in SyncFolderItems.FIELD_ACTION_TYPE_MAP.items():
            for item in val:
                if data.get(item.capitalize()):
                    return Core()._get_recursively(data.get(item.capitalize()).get('ItemId'), '@Id')
                elif data.get(item):
                    return Core()._get_recursively(data.get(item).get('item_id'), 'id')

    def __update_nested(self, in_dict, key):
        for k, v in in_dict.items():
            if key == k:
                details_list = []
                response = SyncFolderItems(k).run()
                for item in response:
                    if item and isinstance(item.get('create'), list):
                        for i in item.get('create'):
                            response = self.__extract_item_id(i)
                            if response:
                                for id in response:
                                    for sync_item in self.__get_item(id):
                                        details_list.append(
                                            self.__dict_merge(Core()._process_dict(i), Core()._process_dict(sync_item))
                                        )
                    elif item and isinstance(item.get('create'), dict):
                        response = self.__extract_item_id(item.get('create'))
                        if response:
                            for id in response:
                                for sync_item in self.__get_item(id):
                                    details_list.append(
                                        self.__dict_merge(Core()._process_dict(item), Core()._process_dict(sync_item))
                                    )
                if self.__folder_hierarchy_details.get(k):
                    in_dict[k].update(self.__folder_hierarchy_details[k])
                if details_list:
                    in_dict[k].update({'details': details_list})
            elif isinstance(v, dict):
                self.__update_nested(v, key)
            elif isinstance(v, list):
                for o in v:
                    if isinstance(o, dict):
                        self.__update_nested(o, key)
        return in_dict

    def sync(self):
        self.__folder_hierarchy_details = {}
        folder_hierarchy_list = []
        for item in SyncFolderHierarchy().run():
            for i in item.get('create'):
                for key,val in i.items():
                    if val.get('FolderId').get('@Id') not in self.__folder_hierarchy_details:
                        self.__folder_hierarchy_details[val.get('FolderId').get('@Id')] = Core()._process_dict(val)
                    folder_hierarchy_list.append((val.get('ParentFolderId').get('@Id'), val.get('FolderId').get('@Id')))
        parents, children = map(set, zip(*folder_hierarchy_list))
        folder_hierarchy_tree = {p: self.__get_children(p, folder_hierarchy_list) for p in (parents - children)}
        self.__sync_folder_items(folder_hierarchy_tree)
        for item in self.__FOLDER_HIERARCHY:
            self.__update_nested(folder_hierarchy_tree, item)
        return folder_hierarchy_tree



from .getsearchablemailboxes import GetSearchableMailboxes
from .getusersettings import GetUserSettings
from .searchmailboxes import SearchMailboxes
from .resolvenames import ResolveNames
from .deleteitem import DeleteItem
from .executesearch import ExecuteSearch
from .getinboxrules import GetInboxRules
from .getitem import GetItem
from .convertid import ConvertId
from .gethiddeninboxrules import GetHiddenInboxRules
from .getattachment import GetAttachment
from .syncfolderhierarchy import SyncFolderHierarchy
from .syncfolderitems import SyncFolderItems
from .createitem import CreateItem
from .getserviceconfiguration import GetServiceConfiguration
from .getdomainsettings import GetDomainSettings
from .createfolder import CreateFolder
from .deletefolder import DeleteFolder
from .finditem import FindItem
from .findfolder import FindFolder
from .deletefolder import DeleteFolder


import uuid
from ..service import Operation


class ExecuteSearch(Operation):
    """ExecuteSearch EWS Operation executes a search as an Outlook client.
    """

    RESULTS_KEY = 'Items'

    def __init__(self, query, result_row_count='25', max_results_count='-1'):
        """Executes a search as an Outlook client for the authenticated user.

        Args:
            query (str): The query to search for.
            result_row_count (str, optional): The row count of results. Defaults to '25'.
            max_results_count (str, optional): The max results count. -1 equals unlimited. Defaults to '-1'.
        """
        self.query = query
        self.result_row_count = result_row_count
        self.max_results_count = max_results_count

    def soap(self):
        self.session_id = '{id}'.format(id=str(uuid.uuid4()).upper())
        return self.M_NAMESPACE.ExecuteSearch(
            self.M_NAMESPACE.ApplicationId('Outlook'),
            self.M_NAMESPACE.Scenario('MailSearch'),
            self.M_NAMESPACE.SearchSessionId(self.session_id),
            self.M_NAMESPACE.SearchScope(
                self.T_NAMESPACE.PrimaryMailboxSearchScope(
                    self.T_NAMESPACE.IsDeepTraversal('true')
                )
            ),
            self.M_NAMESPACE.Query(self.query),
            self.M_NAMESPACE.ResultRowCount(self.result_row_count),
            self.M_NAMESPACE.MaxResultsCountHint(self.max_results_count),
            self.M_NAMESPACE.ItemTypes('MailItems')
        )



from .base import Base, ElementMaker, abc, etree, Authentication


class Autodiscover(Base):
    """Autodiscover class inherits from the Base class
    and defines namespaces, headers, and body
    of an Autodiscover SOAP request
    """

    AUTODISCOVER_MAP  = {
        'wsa': "http://www.w3.org/2005/08/addressing",
        'xsi': "http://www.w3.org/2001/XMLSchema-instance",
        'soap': "http://schemas.xmlsoap.org/soap/envelope/",
        'a': "http://schemas.microsoft.com/exchange/2010/Autodiscover"
    }
    AUTODISCOVER_NAMESPACE = ElementMaker(
        namespace=Base.NAMESPACE_MAP['soap'],
        nsmap=AUTODISCOVER_MAP
    )
    BODY_ELEMENT = ElementMaker(namespace=AUTODISCOVER_MAP['soap']).Body
    A_NAMESPACE = ElementMaker(namespace=AUTODISCOVER_MAP['a'], nsmap={'a': AUTODISCOVER_MAP['a']})
    WSA_NAMESPACE = ElementMaker(namespace=AUTODISCOVER_MAP['wsa'], nsmap={'wsa': AUTODISCOVER_MAP['wsa']})

    @property
    def to(self):
        return Authentication.credentials[0]

    def get(self, exchange_version):
        self.__logger.info('Building Autodiscover SOAP request for {current}'.format(current=self.__class__.__name__))
        ENVELOPE = self.AUTODISCOVER_NAMESPACE.Envelope
        HEADER = self.SOAP_NAMESPACE.Header   
        self.A_NAMESPACE = ElementMaker(namespace=self.AUTODISCOVER_MAP['a'], nsmap={'a': self.AUTODISCOVER_MAP['a']})
        self.envelope = ENVELOPE(
            HEADER(
                self.A_NAMESPACE.RequestedServerVersion(exchange_version),
                self.WSA_NAMESPACE.Action('http://schemas.microsoft.com/exchange/2010/Autodiscover/Autodiscover/{}'.format(self.__class__.__name__)),
                self.WSA_NAMESPACE.To(self.to)
            ),
            self.BODY_ELEMENT(
                self.soap()
            )
        )
        return etree.tostring(self.envelope)

    @abc.abstractmethod
    def soap(self):
        raise NotImplementedError



from ..service import Operation


class SyncFolderItems(Operation):
    """SyncFolderItems EWS Operation retrieves details about folder items
    """

    RESULTS_KEY = 'Changes'
    FIELD_ACTION_TYPE_MAP = {
        'create': [
            'Item',
            'Message',
            'CalendarItem',
            'Contact',
            'DistributionList',
            'MeetingMessage',
            'MeetingRequest',
            'MeetingResponse',
            'MeetingCancellation',
            'Task'
        ],
        'update': [
            'Item',
            'Message',
            'CalendarItem',
            'Contact',
            'DistributionList',
            'MeetingMessage',
            'MeetingRequest',
            'MeetingResponse',
            'MeetingCancellation',
            'Task'
        ],
        'delete': [
            'ItemId'
        ],
        'readflagchange': [
            'ItemId',
            'IsRead'
        ]
    }
    FIELD_URI_PROPERTIES = [
        "folder:FolderId",
        "folder:ParentFolderId",
        "folder:DisplayName",
        "folder:UnreadCount",
        "folder:TotalCount",
        "folder:ChildFolderCount",
        "folder:FolderClass",
        "folder:SearchParameters",
        "folder:ManagedFolderInformation",
        "folder:PermissionSet",
        "folder:EffectiveRights",
        "folder:SharingEffectiveRights",
        "item:ItemId",
        "item:ParentFolderId",
        "item:ItemClass",
        "item:MimeContent",
        "item:Attachments",
        "item:Subject",
        "item:DateTimeReceived",
        "item:Size",
        "item:Categories",
        "item:HasAttachments",
        "item:Importance",
        "item:InReplyTo",
        "item:InternetMessageHeaders",
        "item:IsAssociated",
        "item:IsDraft",
        "item:IsFromMe",
        "item:IsResend",
        "item:IsSubmitted",
        "item:IsUnmodified",
        "item:DateTimeSent",
        "item:DateTimeCreated",
        "item:Body",
        "item:ResponseObjects",
        "item:Sensitivity",
        "item:ReminderDueBy",
        "item:ReminderIsSet",
        "item:ReminderNextTime",
        "item:ReminderMinutesBeforeStart",
        "item:DisplayTo",
        "item:DisplayCc",
        "item:Culture",
        "item:EffectiveRights",
        "item:LastModifiedName",
        "item:LastModifiedTime",
        "item:ConversationId",
        "item:UniqueBody",
        "item:Flag",
        "item:StoreEntryId",
        "item:InstanceKey",
        "item:NormalizedBody",
        "item:EntityExtractionResult",
        "itemPolicyTag",
        "item:ArchiveTag",
        "item:RetentionDate",
        "item:Preview",
        "item:NextPredictedAction",
        "item:GroupingAction",
        "item:PredictedActionReasons",
        "item:IsClutter",
        "item:RightsManagementLicenseData",
        "item:BlockStatus",
        "item:HasBlockedImages",
        "item:WebClientReadFormQueryString",
        "item:WebClientEditFormQueryString",
        "item:TextBody",
        "item:IconIndex",
        "item:MimeContentUTF8",
        "message:ConversationIndex",
        "message:ConversationTopic",
        "message:InternetMessageId",
        "message:IsRead",
        "message:IsResponseRequested",
        "message:IsReadReceiptRequested",
        "message:IsDeliveryReceiptRequested",
        "message:ReceivedBy",
        "message:ReceivedRepresenting",
        "message:References",
        "message:ReplyTo",
        "message:From",
        "message:Sender",
        "message:ToRecipients",
        "message:CcRecipients",
        "message:BccRecipients",
        "message:ApprovalRequestData",
        "message:VotingInformation",
        "message:ReminderMessageData",
        "meeting:AssociatedCalendarItemId",
        "meeting:IsDelegated",
        "meeting:IsOutOfDate",
        "meeting:HasBeenProcessed",
        "meeting:ResponseType",
        "meeting:ProposedStart",
        "meeting:PropsedEnd",
        "meetingRequest:MeetingRequestType",
        "meetingRequest:IntendedFreeBusyStatus",
        "meetingRequest:ChangeHighlights",
        "calendar:Start",
        "calendar:End",
        "calendar:OriginalStart",
        "calendar:StartWallClock",
        "calendar:EndWallClock",
        "calendar:StartTimeZoneId",
        "calendar:EndTimeZoneId",
        "calendar:IsAllDayEvent",
        "calendar:LegacyFreeBusyStatus",
        "calendar:Location",
        "calendar:When",
        "calendar:IsMeeting",
        "calendar:IsCancelled",
        "calendar:IsRecurring",
        "calendar:MeetingRequestWasSent",
        "calendar:IsResponseRequested",
        "calendar:CalendarItemType",
        "calendar:MyResponseType",
        "calendar:Organizer",
        "calendar:RequiredAttendees",
        "calendar:OptionalAttendees",
        "calendar:Resources",
        "calendar:ConflictingMeetingCount",
        "calendar:AdjacentMeetingCount",
        "calendar:ConflictingMeetings",
        "calendar:AdjacentMeetings",
        "calendar:Duration",
        "calendar:TimeZone",
        "calendar:AppointmentReplyTime",
        "calendar:AppointmentSequenceNumber",
        "calendar:AppointmentState",
        "calendar:Recurrence",
        "calendar:FirstOccurrence",
        "calendar:LastOccurrence",
        "calendar:ModifiedOccurrences",
        "calendar:DeletedOccurrences",
        "calendar:MeetingTimeZone",
        "calendar:ConferenceType",
        "calendar:AllowNewTimeProposal",
        "calendar:IsOnlineMeeting",
        "calendar:MeetingWorkspaceUrl",
        "calendar:NetShowUrl",
        "calendar:UID",
        "calendar:RecurrenceId",
        "calendar:DateTimeStamp",
        "calendar:StartTimeZone",
        "calendar:EndTimeZone",
        "calendar:JoinOnlineMeetingUrl",
        "calendar:OnlineMeetingSettings",
        "calendar:IsOrganizer",
        "task:ActualWork",
        "task:AssignedTime",
        "task:BillingInformation",
        "task:ChangeCount",
        "task:Companies",
        "task:CompleteDate",
        "task:Contacts",
        "task:DelegationState",
        "task:Delegator",
        "task:DueDate",
        "task:IsAssignmentEditable",
        "task:IsComplete",
        "task:IsRecurring",
        "task:IsTeamTask",
        "task:Mileage",
        "task:Owner",
        "task:PercentComplete",
        "task:Recurrence",
        "task:StartDate",
        "task:Status",
        "task:StatusDescription",
        "task:TotalWork",
        "contacts:Alias",
        "contacts:AssistantName",
        "contacts:Birthday",
        "contacts:BusinessHomePage",
        "contacts:Children",
        "contacts:Companies",
        "contacts:CompanyName",
        "contacts:CompleteName",
        "contacts:ContactSource",
        "contacts:Culture",
        "contacts:Department",
        "contacts:DisplayName",
        "contacts:DirectoryId",
        "contacts:DirectReports",
        "contacts:EmailAddresses",
        "contacts:FileAs",
        "contacts:FileAsMapping",
        "contacts:Generation",
        "contacts:GivenName",
        "contacts:ImAddresses",
        "contacts:Initials",
        "contacts:JobTitle",
        "contacts:Manager",
        "contacts:ManagerMailbox",
        "contacts:MiddleName",
        "contacts:Mileage",
        "contacts:MSExchangeCertificate",
        "contacts:Nickname",
        "contacts:Notes",
        "contacts:OfficeLocation",
        "contacts:PhoneNumbers",
        "contacts:PhoneticFullName",
        "contacts:PhoneticFirstName",
        "contacts:PhoneticLastName",
        "contacts:Photo",
        "contacts:PhysicalAddresses",
        "contacts:PostalAddressIndex",
        "contacts:Profession",
        "contacts:SpouseName",
        "contacts:Surname",
        "contacts:WeddingAnniversary",
        "contacts:UserSMIMECertificate",
        "contacts:HasPicture",
        "distributionlist:Members",
        "postitem:PostedTime",
        "conversation:ConversationId",
        "conversation:ConversationTopic",
        "conversation:UniqueRecipients",
        "conversation:GlobalUniqueRecipients",
        "conversation:UniqueUnreadSenders",
        "conversation:GlobalUniqueUnreadSenders",
        "conversation:UniqueSenders",
        "conversation:GlobalUniqueSenders",
        "conversation:LastDeliveryTime",
        "conversation:GlobalLastDeliveryTime",
        "conversation:Categories",
        "conversation:GlobalCategories",
        "conversation:FlagStatus",
        "conversation:GlobalFlagStatus",
        "conversation:HasAttachments",
        "conversation:GlobalHasAttachments",
        "conversation:HasIrm",
        "conversation:GlobalHasIrm",
        "conversation:MessageCount",
        "conversation:GlobalMessageCount",
        "conversation:UnreadCount",
        "conversation:GlobalUnreadCount",
        "conversation:Size",
        "conversation:GlobalSize",
        "conversation:ItemClasses",
        "conversation:GlobalItemClasses",
        "conversation:Importance",
        "conversation:GlobalImportance",
        "conversation:ItemIds",
        "conversation:GlobalItemIds",
        "conversation:LastModifiedTime",
        "conversation:InstanceKey",
        "conversation:Preview",
        "conversation:GlobalParentFolderId",
        "conversation:NextPredictedAction",
        "conversation:GroupingAction",
        "conversation:IconIndex",
        "conversation:GlobalIconIndex",
        "conversation:DraftItemIds",
        "conversation:HasClutter",
        "persona:PersonaId",
        "persona:PersonaType",
        "persona:GivenName",
        "persona:CompanyName",
        "persona:Surname",
        "persona:DisplayName",
        "persona:EmailAddress",
        "persona:FileAs",
        "persona:HomeCity",
        "persona:CreationTime",
        "persona:RelevanceScore",
        "persona:WorkCity",
        "persona:PersonaObjectStatus",
        "persona:FileAsId",
        "persona:DisplayNamePrefix",
        "persona:YomiCompanyName",
        "persona:YomiFirstName",
        "persona:YomiLastName",
        "persona:Title",
        "persona:EmailAddresses",
        "persona:PhoneNumber",
        "persona:ImAddress",
        "persona:ImAddresses",
        "persona:ImAddresses2",
        "persona:ImAddresses3",
        "persona:FolderIds",
        "persona:Attributions",
        "persona:DisplayNames",
        "persona:Initials",
        "persona:FileAses",
        "persona:FileAsIds",
        "persona:DisplayNamePrefixes",
        "persona:GivenNames",
        "persona:MiddleNames",
        "persona:Surnames",
        "persona:Generations",
        "persona:Nicknames",
        "persona:YomiCompanyNames",
        "persona:YomiFirstNames",
        "persona:YomiLastNames",
        "persona:BusinessPhoneNumbers",
        "persona:BusinessPhoneNumbers2",
        "persona:HomePhones",
        "persona:HomePhones2",
        "persona:MobilePhones",
        "persona:MobilePhones2",
        "persona:AssistantPhoneNumbers",
        "persona:CallbackPhones",
        "persona:CarPhones",
        "persona:HomeFaxes",
        "persona:OrganizationMainPhones",
        "persona:OtherFaxes",
        "persona:OtherTelephones",
        "persona:OtherPhones2",
        "persona:Pagers",
        "persona:RadioPhones",
        "persona:TelexNumbers",
        "persona:WorkFaxes",
        "persona:Emails1",
        "persona:Emails2",
        "persona:Emails3",
        "persona:BusinessHomePages",
        "persona:School",
        "persona:PersonalHomePages",
        "persona:OfficeLocations",
        "persona:BusinessAddresses",
        "persona:HomeAddresses",
        "persona:OtherAddresses",
        "persona:Titles",
        "persona:Departments",
        "persona:CompanyNames",
        "persona:Managers",
        "persona:AssistantNames",
        "persona:Professions",
        "persona:SpouseNames",
        "persona:Hobbies",
        "persona:WeddingAnniversaries",
        "persona:Birthdays",
        "persona:Children",
        "persona:Locations",
        "persona:ExtendedProperties",
        "persona:PostalAddress",
        "persona:Bodies"
    ]

    def __init__(self, folder_id, change_key=None):
        """Retrieves details about a provided folder id.

        Args:
            folder_id (str): The folder id to retrieve details about.
            change_key (str, optional): The version key of the folder id. Defaults to None.
        """
        self.folder_id = folder_id
        self.change_key = change_key

    def soap(self):
        if self.change_key:
            folder_id = self.T_NAMESPACE.FolderId(Id=self.folder_id, ChangeKey=self.change_key)
        else:
            folder_id = self.T_NAMESPACE.FolderId(Id=self.folder_id)
        return self.M_NAMESPACE.SyncFolderItems(
            self.M_NAMESPACE.ItemShape(
                self.T_NAMESPACE.BaseShape('AllProperties'),
                self.T_NAMESPACE.IncludeMimeContent('true'),
                self.T_NAMESPACE.BodyType('Best'),
                self.T_NAMESPACE.FilterHtmlContent('false'),
                self.T_NAMESPACE.ConvertHtmlCodePageToUTF8('false')
            ),
            self.M_NAMESPACE.SyncFolderId(
                folder_id
            ),
            self.M_NAMESPACE.MaxChangesReturned('100'),
            self.M_NAMESPACE.SyncScope('NormalAndAssociatedItems')
        )



class ExchangeVersion:
    '''Used to validate compatible Exchange Versions across multiple service endpoints

    Examples:
        To determine if a version number is a valid ExchangeVersion then would pass in the value when instantiating this object:

        ```python
        version = ExchangeVersion('15.20.5').exchangeVersion
        print(version)
        ```

        ```text
        # output
        Exchange2016
        ```

        To verify an ExchangeVersion is supported, you can view the supported version by access the EXCHANGE_VERSIONS attribute

        ```python
        versions = ExchangeVersion('15.20.5').EXCHANGE_VERSIONS
        print(versions)
        ```

        ```text    
        ['Exchange2019', 'Exchange2016', 'Exchange2013_SP1', 'Exchange2013', 'Exchange2010_SP2', 'Exchange2010_SP1', 'Exchange2010']
        ```

    Args:
        version (str): An Exchange Version number.  Example: 15.20.5 = Exchange2016
    '''

    # Borrowed from exchangelib: https://github.com/ecederstrand/exchangelib/blob/master/exchangelib/version.py#L54
    # List of build numbers here: https://docs.microsoft.com/en-us/Exchange/new-features/build-numbers-and-release-dates?view=exchserver-2019 
    API_VERSION_MAP = {
        8: {
            0: 'Exchange2007',
            1: 'Exchange2007_SP1',
            2: 'Exchange2007_SP1',
            3: 'Exchange2007_SP1',
        },
        14: {
            0: 'Exchange2010',
            1: 'Exchange2010_SP1',
            2: 'Exchange2010_SP2',
            3: 'Exchange2010_SP2',
        },
        15: {
            0: 'Exchange2013',  # Minor builds starting from 847 are Exchange2013_SP1, see api_version()
            1: 'Exchange2016',
            2: 'Exchange2019',
            20: 'Exchange2016',  # This is Office365. See issue #221
        }
    }

    EXCHANGE_VERSIONS = ['Exchange2019', 'Exchange2016', 'Exchange2015', 'Exchange2013_SP1', 'Exchange2013', 'Exchange2010_SP2', 'Exchange2010_SP1', 'Exchange2010']

    def __init__(self, version):
        self.exchange_version = self._get_api_version(version)

    def _get_api_version(self, version):
        '''Gets a string representation of an Exchange Version number

        Args:
            version (str): An Exchange Version number.  Example: 15.20.5

        Returns:
            str: A string representation of a Exchange Version number. Example: Exchange2016
        '''

        if version == '15.0.847.32':
            return 'Exchange2013_SP1'
        elif version == '15.20.3955.27':
            return 'Exchange2015'
        else:
            ver = version.split('.')
            return self.API_VERSION_MAP[int(ver[0])][int(ver[1])]

    @staticmethod
    def valid_version(version):
        '''Determines if a string version name is in list of accepted Exchange Versions

        Args:
            version (str): String used to determine if it is an acceptable Exchange Version

        Returns:
            bool: Returns either True or False if the passed in version is an acceptable Exchange Version
        '''
        if version == 'Office365':
            return True
        elif version in ExchangeVersion.EXCHANGE_VERSIONS:
            return True
        return False



class Endpoints:
    """Endpoints provides and generates endpoints to attempt
    EWS requests against.
    """

    def __init__(self, domain=None):
        self.domain = domain

    def get(self):
        endpoint_list = ['https://outlook.office365.com/autodiscover/autodiscover.svc', 'https://outlook.office365.com/EWS/Exchange.asmx', 'https://autodiscover-s.outlook.com/autodiscover/autodiscover.svc']
        if self.domain:
            endpoint_list.append("https://{}/autodiscover/autodiscover.svc".format(self.domain))
            endpoint_list.append("https://autodiscover.{}/autodiscover/autodiscover.svc".format(self.domain))
        return endpoint_list



from ..service.base import Base, etree
from ..utils.exceptions import UknownValueError
from ..utils.attributes import FIELD_URI_MAP, SEARCH_FILTERS


class SearchFilter:
    """SearchFilter is used to generate a search filter XML body for
    the FindItem EWS Operation
    """

    FIELD_NAMES = FIELD_URI_MAP
    EXPRESSIONS = list(SEARCH_FILTERS.keys())

    def __init__(self, value, to_string=False):
        """Pass in a string value and in return this class returns a
        XML object

        Detailed Information:
            Each search filter is made of several elements. These are:
                Field Name - Retrieve all available Field Names by using SearchFilter.FIELD_NAMES
                Expression - Retrieve all expressions by using SearchFilter.EXPRESSIONS
                Value - Values are typically strings which make sens for field names

            A logic operator within the SearchFilter class is one of:
                and - This will perform a boolean AND search operation between two or more search expressions
                or - This will perform a logical OR search operation between two or more search expressions
                not - This will perform a search expression that negates a search expression


        Examples:
            SearchFilter('Body contains Hello World')
            SearchFilter('Subject contains phishing and IsRead IsEqualTo true')
            SearchFilter('Subject contains phishing and IsRead IsEqualTo true or Subject contains microsoft', to_string=True)
        Args:
            value ([str]): A query string
        """
        self.to_string = to_string
        self.__parent = Base.T_NAMESPACE.Restriction()
        self.__build_return_object(value)
        if to_string:
            self._value = str(etree.tostring(self.__parent).decode("utf-8"))
        else:
            self._value = self.__parent

    def __new__(cls, value, to_string=False):
        instance = super(SearchFilter, cls).__new__(cls)
        if hasattr(cls, '__init__'):
            instance.__init__(value, to_string=to_string)
        return instance._value

    @property
    def search_string(self):
        if self.to_string:
            return str(etree.tostring(self.__parent).decode("utf-8"))
        return self.__parent

    def __build_return_object(self, value):
        if 'and' in value:
            object_list = []
            for item in value.split('and'):
                obj = self.__evaluate(item)
                if obj is not None:
                    object_list.append(obj)
            self.__parent.append(Base.T_NAMESPACE.And(*object_list))
        if 'or' in value:
            object_list = []
            for item in value.split('or'):
                obj = self.__evaluate(item)
                if obj is not None:
                    object_list.append(obj)
            self.__parent.append(Base.T_NAMESPACE.Or(*object_list))
        else:
            self.__parent.append(self.__evaluate(value))
        self.__parent

    def __build_search_filter(self, search_filter, field, field_value):
        if search_filter == 'Contains':
            return Base.T_NAMESPACE.Contains(
                Base.T_NAMESPACE.FieldURI(FieldURI=field),
                Base.T_NAMESPACE.FieldURIOrConstant(
                    Base.T_NAMESPACE.Constant(Value=field_value)
                ),
              #  Base.T_NAMESPACE.Constant(Value=field_value),
                ContainmentMode='Substring',
                ContainmentComparison='IgnoreCase'
            )
        elif search_filter == 'Excludes':
            return Base.T_NAMESPACE.Excludes(
                Base.T_NAMESPACE.ExtendedFieldURI(
                    PropertySetId="aa3df801-4fc7-401f-bbc1-7c93d6498c2e",
                    PropertyName="ItemIndex",
                    PropertyType="Integer"
                ),
                Base.T_NAMESPACE.Bitmask(Value=field_value)
            )
        elif search_filter == 'Exists':
            return Base.T_NAMESPACE.Exists(
                Base.T_NAMESPACE.ExtendedFieldURI(
                    PropertySetId="aa3df801-4fc7-401f-bbc1-7c93d6498c2e",
                    PropertyName="ItemIndex",
                    PropertyType="Integer"
                ),
            )
        elif search_filter == 'IsEqualTo':
            return Base.T_NAMESPACE.IsEqualTo(
                Base.T_NAMESPACE.FieldURI(FieldURI=field),
                Base.T_NAMESPACE.FieldURIOrConstant(
                    Base.T_NAMESPACE.Constant(Value=field_value)
                )
            )
        elif search_filter == 'IsNotEqualTo':
            return Base.T_NAMESPACE.IsNotEqualTo(
                Base.T_NAMESPACE.ExtendedFieldURI(
                    PropertySetId="aa3df801-4fc7-401f-bbc1-7c93d6498c2e",
                    PropertyName="ItemIndex",
                    PropertyType="Integer"
                ),
                Base.T_NAMESPACE.FieldURIOrConstant(
                    Base.T_NAMESPACE.FieldURI(FieldURI=field)
                )
            )

    def __to_title_case(self, value):
        return ''.join([x for x in value.title() if not x.isspace()])

    def __evaluate(self, value):
        search_filter = None
        field = None
        field_value = None
        temp_value = value
        for key,val in SEARCH_FILTERS.items():
            if key in value:
                search_filter = key
                temp_value = value.split(key)
                field = temp_value[0].strip()
                field_value = temp_value[1].strip()
            elif key.lower() in value:
                search_filter = key
                temp_value = value.split(key.lower())
                field = temp_value[0].strip()
                field_value = temp_value[1].strip()
        if field:
            for key,val in FIELD_URI_MAP.items():
                if field in val:
                    field = f"{key}:{field}"
                elif self.__to_title_case(field) in val:
                    field = f"{key}:{self.__to_title_case(field)}"
        if ' or ' in field_value:
            field_value = field_value.split(' or ')[0]
        elif ' and ' in field_value:
            field_value = field_value.split(' and ')[0]
        return self.__build_search_filter(search_filter, field, field_value)



from ..service import Operation
from ..utils.exceptions import UknownValueError


class GetItem(Operation):
    """GetItem EWS Operation retrieves details about an item.
    """

    RESULTS_KEY = 'Items'
    BASE_SHAPES = [
        'IdOnly',
        'Default',
        'AllProperties'
    ]
    BODY_TYPES = [
        'Best',
        'HTML',
        'Text'
    ]

    def __init__(self, item_id, change_key=None, base_shape='AllProperties', include_mime_content=True, body_type='Best'):
        """Retrieves details about a provided item id

        Args:
            item_id (str): The item id you want to get information about.
            change_key (str, optional): The change key of the item. Defaults to None.
            base_shape (str, optional): The base shape of the returned item. Defaults to 'AllProperties'.
            include_mime_content (bool, optional): Whether or not to include MIME content. Defaults to True.
            body_type (str, optional): The item body type. Defaults to 'Best'.
        """
        self.include_mime_content = include_mime_content
        if base_shape not in self.BASE_SHAPES:
            UknownValueError(provided_value=base_shape, known_values=self.BASE_SHAPES)
        self.base_shape = base_shape
        if body_type not in self.BODY_TYPES:
            UknownValueError(provided_value=body_type, known_values=self.BODY_TYPES)
        self.body_type = body_type
        self.item_id = item_id
        self.change_key = change_key

    def soap(self):
        if self.change_key:
            item_id_string = self.T_NAMESPACE.ItemId(Id=self.item_id, ChangeKey=self.change_key)
        else:
            item_id_string = self.T_NAMESPACE.ItemId(Id=self.item_id)
        return self.M_NAMESPACE.GetItem(
            self.M_NAMESPACE.ItemShape(
                self.T_NAMESPACE.BaseShape(self.base_shape),
                self.T_NAMESPACE.IncludeMimeContent(str(self.include_mime_content).lower()),
                self.T_NAMESPACE.BodyType(self.body_type)
            ),
            self.M_NAMESPACE.ItemIds(
                item_id_string
            )
        )



from ..service import Operation
from ..utils.exceptions import UknownValueError
from ..utils.attributes import FOLDER_LIST, MESSAGE_ELEMENTS


class AddDelegate(Operation):
    """AddDelegate EWS Operation adds one or more delegates
    to a principal's mailbox and sets access permissions.
    """

    RESULTS_KEY = 'Items'

    __PERMISSIONS = [
        'None',
        'Reviewer',
        'Author',
        'Editor'
    ]

    def __init__(self, target_mailbox, delegate_to, inbox_permissions=None, calender_permissions=None, contacts_permissions=None):
        """Adds the delegate_to to the provided target_mailbox with specific permission sets.

        Args:
            target_mailbox (str): The mailbox in which permissions are added
            delegate_to (str): The account that the permissions are assocaited with
            inbox_permissions (str, optional): Mailbox permission set value. Defaults to None.
            calender_permissions (str, optional): Calendar permission set value. Defaults to None.
            contacts_permissions (str, optional): Contacts permission set value. Defaults to None.
        """
        self.target_mailbox = target_mailbox
        self.delegate_to = delegate_to
        if inbox_permissions and inbox_permissions not in self.__PERMISSIONS:
            UknownValueError(provided_value=inbox_permissions, known_values=self.__PERMISSIONS)
        self.inbox_permissions = inbox_permissions if isinstance(inbox_permissions, list) else [inbox_permissions]

        if calender_permissions and calender_permissions not in self.__PERMISSIONS:
            UknownValueError(provided_value=calender_permissions, known_values=self.__PERMISSIONS)
        self.calender_permissions = calender_permissions if isinstance(calender_permissions, list) else [calender_permissions]

        if contacts_permissions and contacts_permissions not in self.__PERMISSIONS:
            UknownValueError(provided_value=contacts_permissions, known_values=self.__PERMISSIONS)
        self.contacts_permissions = contacts_permissions if isinstance(contacts_permissions, list) else [contacts_permissions]

    def __get_delegate_users(self):
        return_list = []
        if self.inbox_permissions:
            for item in self.inbox_permissions:
                return_list.append(
                    self.T_NAMESPACE.InboxFolderPermissionLevel(item)
                )
        if self.contacts_permissions:
            for item in self.contacts_permissions:
                return_list.append(
                    self.T_NAMESPACE.ContactsFolderPermissionLevel(item)
                )
        if self.calender_permissions:
            for item in self.calender_permissions:
                return_list.append(
                    self.T_NAMESPACE.CalendarFolderPermissionLevel(item)
                )
        return self.T_NAMESPACE.DelegateUser(
            self.T_NAMESPACE.UserId(
                self.T_NAMESPACE.PrimarySmtpAddress(self.delegate_to)
            ),
            self.T_NAMESPACE.DelegatePermissions(
                *return_list
            ),
            self.T_NAMESPACE.ReceiveCopiesOfMeetingMessages('false'),
            self.T_NAMESPACE.ViewPrivateItems('false')
        )

    def soap(self):
        return self.M_NAMESPACE.AddDelegate(
            self.M_NAMESPACE.Mailbox(
                self.T_NAMESPACE.EmailAddress(self.target_mailbox)
            ),
            self.M_NAMESPACE.DelegateUsers(
                self.__get_delegate_users()
            ),
            self.M_NAMESPACE.ReceiveCopiesOfMeetingMessages('DelegatesAndMe')
        )



from ..service import Operation
from ..utils.exceptions import UknownValueError


class DeleteItem(Operation):
    """DeleteItem EWS Operation deletes items in the Exchange store.
    """

    DELETE_TYPES = ['HardDelete', 'SoftDelete', 'MoveToDeletedItems']

    def __init__(self, item_id, delete_type='MoveToDeletedItems'):
        """Deletes the provided Item ID from an Exchange store.

        Args:
            item_id (str): The Item ID to delete
            delete_type (str, optional): The delete type when deleting the item. Defaults to 'MoveToDeletedItems'.
        """
        if not isinstance(item_id, list):
            item_id = [item_id]
        self.item_id = item_id
        if delete_type and delete_type not in self.DELETE_TYPES:
            UknownValueError(provided_value=delete_type, known_values=self.DELETE_TYPES)
        self.delete_type = delete_type

    def soap(self):
        item_id_list = []
        for item in self.item_id:
            item_id_list.append(self.T_NAMESPACE.ItemId(Id=item))
        return self.M_NAMESPACE.DeleteItem(
            self.M_NAMESPACE.ItemIds(
                item_id_list
            ),
            **{'DeleteType':self.delete_type}
        )



import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from .core import Authentication
from .endpoint import GetSearchableMailboxes, GetUserSettings, ResolveNames, SearchMailboxes, ExecuteSearch, GetInboxRules, GetItem, ConvertId, GetHiddenInboxRules, CreateItem, GetServiceConfiguration, SyncFolderHierarchy, SyncFolderItems, GetAttachment, DeleteItem, GetDomainSettings, FindItem, CreateFolder, FindFolder, DeleteFolder


class EWS:

    def __init__(self, 
        username, password, ews_url=None, exchange_version=None, impersonate_as=None, multi_threading=False, 
        tenant_id=None, client_id=None, client_secret=None, oauth2_authorization_type='auth_code_grant', redirect_uri=None, oauth2_scope=None):
        Authentication.credentials = (username, password)
        Authentication.ews_url = ews_url
        Authentication.exchange_versions = exchange_version
        Authentication.impersonate_as = impersonate_as
        Authentication.tenant_id = tenant_id
        Authentication.client_id = client_id
        Authentication.client_secret = client_secret
        Authentication.oauth2_authorization_type = oauth2_authorization_type
        Authentication.redirect_uri = redirect_uri
        Authentication.oauth2_scope = oauth2_scope
        self.multi_threading = multi_threading

    def chunk(self, items, n):
        n = max(1, n)
        return (items[i:i+n] for i in range(0, len(items), n))

    def get_service_configuration(self, configuration_name=None, acting_as=None):
        return GetServiceConfiguration(configuration_name=configuration_name, acting_as=acting_as).run()

    def get_searchable_mailboxes(self, search_filter=None, expand_group_memberhip=True):
        return GetSearchableMailboxes(search_filter=search_filter, expand_group_memberhip=expand_group_memberhip).run()

    def get_user_settings(self, user=None):
        return GetUserSettings(user=user).run()

    def resolve_names(self, user=None):
        return ResolveNames(user=user).run()

    def __execute_multithreaded_search(self, query, reference_id, scope):
        return SearchMailboxes(query=query, reference_id=reference_id, search_scope=scope).run()

    def __execute_multithreaded_find(
            self, 
            query_string, 
            user_list,
            distinguished_folder_name='inbox', 
            base_shape='AllProperties', 
            include_mime_content=True, 
            body_type='Best', 
            traversal='Shallow', 
            reset_cache=False, 
            return_deleted_items=True, 
            return_highlight_terms=True
        ):
        return_list = []
        for user in user_list:
            if user:
                Authentication.impersonate_as = user
                return_list.extend(FindItem(query_string=query_string, distinguished_folder_name=distinguished_folder_name, base_shape=base_shape, include_mime_content=include_mime_content, body_type=body_type, traversal=traversal, reset_cache=reset_cache, return_deleted_items=return_deleted_items, return_highlight_terms=return_highlight_terms).run())
        return return_list

    def execute_ews_search(self, query, reference_id, search_scope='All', thread_count=os.cpu_count()*5):
        response = []
        return_list = []
        if self.multi_threading:
            threads = []
            chunks = self.chunk(reference_id, int(len(reference_id) / thread_count))
            with ThreadPoolExecutor(max_workers=thread_count) as executor:
                for chunk in chunks:
                    threads.append(executor.submit(self.__execute_multithreaded_search, query, chunk, search_scope))
                for task in as_completed(threads):
                    result = task.result()
                    if isinstance(result, list):
                        for item in result:
                            response.append(item)
        else:
            response = SearchMailboxes(query, reference_id=reference_id, search_scope=search_scope).run()
        for item in response:
            if item:
                return_dict = item
                get_item_response = self.get_item(return_dict['id'].get('id'))
                if get_item_response:
                    for item_response in get_item_response:
                        if item_response and item_response.get('message').get('attachments') and item_response.get('message').get('attachments').get('file_attachment').get('attachment_id').get('id'):
                            attachment_details_list = []
                            attachment = self.get_attachment(item_response.get('message').get('attachments').get('file_attachment').get('attachment_id').get('id'))
                            for attach in attachment:
                                attachment_dict = {}
                                if attach:
                                    for key,val in attach.items():
                                        for k,v in val.items():
                                            attachment_dict[k] = v
                                if attachment_dict:
                                    attachment_details_list.append(attachment_dict)
                            if attachment_details_list:
                                return_dict.update({'attachment_details': attachment_details_list})
                        return_dict.update(item_response.pop('message'))
                return_list.append(return_dict)
        return return_list

    def execute_outlook_search(self, query, result_row_count='25', max_results_count='-1'):
        return ExecuteSearch(
                query=query, 
                result_row_count=result_row_count, 
                max_results_count=max_results_count
            ).run()

    def get_inbox_rules(self, user=None):
        return GetInboxRules(user=user).run()

    def get_hidden_inbox_rules(self):
        return GetHiddenInboxRules().run()

    def get_item(self, item_id, change_key=None):
        response = GetItem(item_id, change_key=change_key).run()
        if isinstance(response, list) and response:
            if any(item in response for item in ConvertId.ID_FORMATS):
                convert_id_response = ConvertId(Authentication.credentials[0], item_id, id_type=response[0], convert_to=response[1]).run()
                get_item_response = GetItem(convert_id_response[0]).run()
                return get_item_response if get_item_response else None
        return GetItem(item_id, change_key=change_key).run()

    def get_attachment(self, attachment_id):
        return GetAttachment(attachment_id=attachment_id).run()

    def sync_folder_hierarchy(self, well_known_folder_name=None):
        return SyncFolderHierarchy(well_known_folder_name=well_known_folder_name).run()

    def sync_folder_items(self, folder_id, change_key=None):
        return SyncFolderItems(folder_id, change_key=change_key).run()

    def create_item(self, subject, sender, to_recipients, body_type='HTML'):
        return CreateItem(**{'Subject': subject, 'BodyType': body_type, 'Sender': sender, 'ToRecipients': to_recipients}).run()

    def delete_item(self, item_id, delete_type='MoveToDeletedItems'):
        if not isinstance(item_id, list):
            item_id = [item_id]
        return_list = []
        for item in item_id:
            if item:
                return_list.append(DeleteItem(item, delete_type=delete_type).run())
        return return_list

    def search_and_delete_message(self, query, thread_count=os.cpu_count()*5, what_if=False):
        return_list = []
        reference_id_list = []
        for mailbox in self.get_searchable_mailboxes():
            if mailbox:
                reference_id_list.append(mailbox.get('reference_id'))
        if not reference_id_list:
            print('No searchable mailboxes found')
            return None
        else:
            for item in self.execute_ews_search(query, reference_id_list, thread_count=thread_count):
                if item:
                    if what_if:
                        print('WHAT IF: About to delete message ID: {}'.format(item.get('id').get('id')))
                    else:
                        return_list.extend(self.delete_item(item.get('id').get('id')))
        return return_list

    def get_domain_settings(self, domain=None):
        return GetDomainSettings(domain=domain).run()

    def find_items(self, 
                    query_string, 
                    distinguished_folder_name='inbox', 
                    base_shape='AllProperties', 
                    include_mime_content=True, 
                    body_type='Best', 
                    traversal='Shallow', 
                    reset_cache=False, 
                    return_deleted_items=True, 
                    return_highlight_terms=True
        ):
        return_list = []
        for message in FindItem(query_string=query_string, distinguished_folder_name=distinguished_folder_name, base_shape=base_shape, include_mime_content=include_mime_content, body_type=body_type, traversal=traversal, reset_cache=reset_cache, return_deleted_items=return_deleted_items, return_highlight_terms=return_highlight_terms).run():
            if message:
                items = self.get_item(item_id=message.get('item_id').get('id'))
                if items:
                    for item in items:
                        return_list.append(item)
        return return_list

    def search_mailboxes_using_find_item(self, 
                    query_string, 
                    impersonation_list,
                    multi_threading=True,
                    thread_count=os.cpu_count()*5,
                    distinguished_folder_name='inbox', 
                    base_shape='AllProperties', 
                    include_mime_content=True, 
                    body_type='Best', 
                    traversal='Shallow', 
                    reset_cache=False, 
                    return_deleted_items=True, 
                    return_highlight_terms=True):
        if multi_threading:
            threads = []
            response = []
            chunks = self.chunk(impersonation_list, int(len(impersonation_list) / thread_count))
            with ThreadPoolExecutor(max_workers=thread_count) as executor:
                for chunk in chunks:
                    threads.append(executor.submit(self.__execute_multithreaded_find, query_string, distinguished_folder_name, base_shape, include_mime_content, body_type, traversal, reset_cache, return_deleted_items, return_highlight_terms, chunk))
                for task in as_completed(threads):
                    result = task.result()
                    if isinstance(result, list):
                        for item in result:
                            response.append(item)
            return response

    def create_search_folder(self, search_string, display_name='Search Folder', base_folder='inbox', traversal='Deep'):
        return CreateFolder(search_string=search_string, search_folder=True, display_name=display_name, base_folder=base_folder, traversal=traversal).run()

    def find_search_folder(self):
        return FindFolder().run()

    def delete_search_folder(self, folder_id):
        return DeleteFolder(folder_id=folder_id)



import xmltodict
import json

from ..utils.logger import LoggingBase


class Core(metaclass=LoggingBase):
    """The Core class inherits logging and defines
    required authentication details as well as parsing of results
    """

    def camel_to_snake(self, s):
        if s != 'UserDN':
            return ''.join(['_'+c.lower() if c.isupper() else c for c in s]).lstrip('_')
        else:
            return 'user_dn'

    def __process_keys(self, key):
        return_value = key.replace('t:','')
        if return_value.startswith('@'):
            return_value = return_value.lstrip('@')
        return self.camel_to_snake(return_value)

    def _process_dict(self, obj):
        if isinstance(obj, dict):
            obj = {
                self.__process_keys(key): self._process_dict(value) for key, value in obj.items()
                }
        return obj

    def _get_recursively(self, search_dict, field):
        """
        Takes a dict with nested lists and dicts,
        and searches all dicts for a key of the field
        provided.
        """
        fields_found = []
        if search_dict:
            for key, value in search_dict.items():
                if key == field:
                    fields_found.append(value)
                elif isinstance(value, dict):
                    results = self._get_recursively(value, field)
                    for result in results:
                        fields_found.append(result)
                elif isinstance(value, list):
                    for item in value:
                        if isinstance(item, dict):
                            more_results = self._get_recursively(item, field)
                            for another_result in more_results:
                                fields_found.append(another_result)
        return fields_found

    def parse_response(self, soap_response, namespace_dict=None):
        """parse_response is standardized to parse all soap_responses from
        EWS requests

        Args:
            soap_response (BeautifulSoup): EWS SOAP response returned from the Base class
            namespace_dict (dict, optional): A dictionary of namespaces to process. Defaults to None.

        Returns:
            list: Returns a list of dictionaries containing parsed responses from EWS requests.
        """
        ordered_dict = xmltodict.parse(str(soap_response), process_namespaces=True, namespaces=namespace_dict)
        item_dict = json.loads(json.dumps(ordered_dict))
        if hasattr(self, 'RESULTS_KEY'):
            search_response = self._get_recursively(item_dict, self.RESULTS_KEY)
            if search_response:
                return_list = []
                for item in search_response:
                    if isinstance(item,list):
                        for i in item:
                            return_list.append(self._process_dict(i))
                    else:
                        return_list.append(self._process_dict(item))
                return return_list
        return self._process_dict(item_dict)



from ..service import Operation


class GetAttachment(Operation):
    """GetAttachment EWS Operation retrieves the provided attachment ID.
    """

    RESULTS_KEY = 'Attachments'

    def __init__(self, attachment_id):
        """Retrieves the provided attachment_id

        Args:
            attachment_id (str): The attachment_id to retrieve.
        """
        self.attachment_id = attachment_id

    def soap(self):
        return self.M_NAMESPACE.GetAttachment(
            self.M_NAMESPACE.AttachmentShape(
                self.T_NAMESPACE.IncludeMimeContent('true'),
                self.T_NAMESPACE.BodyType('Best')
            ),
            self.M_NAMESPACE.AttachmentIds(
                self.T_NAMESPACE.AttachmentId(Id=self.attachment_id)
            )
        )



from ..service import Operation


class GetSearchableMailboxes(Operation):
    """GetSearchableMailboxes EWS Operation retrieves a
    list of searchable mailboxes the authenticated user has
    access to search.
    """

    RESULTS_KEY = 'SearchableMailbox'

    def __init__(self, search_filter=None, expand_group_memberhip=True):
        """Retrieves all searchable mailboxes the authenticated user has access to search.

        Args:
            search_filter (str, optional): A search filter. Typically used to search specific distribution groups. Defaults to None.
            expand_group_memberhip (bool, optional): Whether or not to expand group memberships. Defaults to True.
        """
        self.search_filter = self.M_NAMESPACE.SearchFilter(search_filter) if search_filter else self.M_NAMESPACE.SearchFilter()
        self.expand_group_membership = self.M_NAMESPACE.ExpandGroundMembership(str(expand_group_memberhip))

    def soap(self):
        return self.M_NAMESPACE.GetSearchableMailboxes(
            self.search_filter,
            self.expand_group_membership
        )



from ..service import Operation


class GetHiddenInboxRules(Operation):
    """FindItem EWS Operation attempts to find
    hidden inbox rules based on ExtendedFieldURI properties.
    """

    RESULTS_KEY = 'Items'

    def soap(self):
        return self.M_NAMESPACE.FindItem(
            self.M_NAMESPACE.ItemShape(
                self.T_NAMESPACE.BaseShape('IdOnly'),
                self.T_NAMESPACE.AdditionalProperties(
                    self.T_NAMESPACE.ExtendedFieldURI(PropertyTag="0x65EC", PropertyType="String"),
                    self.T_NAMESPACE.ExtendedFieldURI(PropertyTag="0x0E99", PropertyType="Binary"),
                    self.T_NAMESPACE.ExtendedFieldURI(PropertyTag="0x0E9A", PropertyType="Binary"),
                    self.T_NAMESPACE.ExtendedFieldURI(PropertyTag="0x65E9", PropertyType="Integer"),
                    self.T_NAMESPACE.ExtendedFieldURI(PropertyTag="0x6800", PropertyType="String"),
                    self.T_NAMESPACE.ExtendedFieldURI(PropertyTag="0x65EB", PropertyType="String"),
                    self.T_NAMESPACE.ExtendedFieldURI(PropertyTag="0x3FEA", PropertyType="Boolean"),
                    self.T_NAMESPACE.ExtendedFieldURI(PropertyTag="0x6645", PropertyType="Binary"),
                )
            ),
            self.M_NAMESPACE.Restriction(
                self.T_NAMESPACE.IsEqualTo(
                    self.T_NAMESPACE.FieldURI(FieldURI="item:ItemClass"),
                    self.T_NAMESPACE.FieldURIOrConstant(
                        self.T_NAMESPACE.Constant(Value="IPM.Rule.Version2.Message")
                    )
                )
            ),
            self.M_NAMESPACE.ParentFolderIds(
                self.T_NAMESPACE.DistinguishedFolderId(Id="inbox")
            ),
            Traversal="Shallow")



from ..service import Operation


class GetInboxRules(Operation):
    """GetInboxRules EWS Operation retrieves inbox rules.
    """

    RESULTS_KEY = 'GetInboxRulesResponse'

    def __init__(self, user=None):
        """Retrieves inbox rules for the authenticated or specified user.

        Args:
            user (str, optional): The user to retrieve inbox rules for. Defaults to authenticated user.
        """
        self.__user = user

    def soap(self):
        if not self.__user:
            self.__user = self.credentials[0]
        return self.M_NAMESPACE.GetInboxRules(
            self.M_NAMESPACE.MailboxSmtpAddress(self.__user)
        )



from .ews import EWS
from .core import Core, ExchangeVersion, Authentication, Endpoints, OAuth2Connector
from .endpoint import *
from .service import Autodiscover, Operation
from .utils.searchfilter import SearchFilter



from ..service.autodiscover import Autodiscover, Authentication


class GetUserSettings(Autodiscover):
    """GetUserSettings EWS Autodiscover endpoint
    retrieves the authenticated or provided users settings
    """

    RESULTS_KEY = 'UserSettings'

    def __init__(self, user=None):
        """Retrieves the user settings for the authenticated or provided user.

        Args:
            user (str, optional): A user to retrieve user settings for. Defaults to None.
        """
        self.user = user

    def soap(self):
        if not self.user:
            self.user = Authentication.credentials[0]
        return self.A_NAMESPACE.GetUserSettingsRequestMessage(
            self.A_NAMESPACE.Request(
                self.A_NAMESPACE.Users(
                    self.A_NAMESPACE.User(
                        self.A_NAMESPACE.Mailbox(self.user)
                    )
                ),
                self.A_NAMESPACE.RequestedSettings(
                    self.A_NAMESPACE.Setting('InternalEwsUrl'),
                    self.A_NAMESPACE.Setting('ExternalEwsUrl'),
                    self.A_NAMESPACE.Setting('UserDisplayName'),
                    self.A_NAMESPACE.Setting('UserDN'),
                    self.A_NAMESPACE.Setting('UserDeploymentId'),
                    self.A_NAMESPACE.Setting('InternalMailboxServer'),
                    self.A_NAMESPACE.Setting('MailboxDN'),
                    self.A_NAMESPACE.Setting('ActiveDirectoryServer'),
                    self.A_NAMESPACE.Setting('EwsSupportedSchemas'),
                    self.A_NAMESPACE.Setting('InternalRpcClientServer'),
                    self.A_NAMESPACE.Setting('InternalEcpUrl'),
                    self.A_NAMESPACE.Setting('InternalEcpVoicemailUrl'),
                    self.A_NAMESPACE.Setting('InternalEcpEmailSubscriptionsUrl'),
                    self.A_NAMESPACE.Setting('InternalEcpTextMessagingUrl'),
                    self.A_NAMESPACE.Setting('InternalEcpDeliveryReportUrl'),
                    self.A_NAMESPACE.Setting('InternalEcpRetentionPolicyTagsUrl'),
                    self.A_NAMESPACE.Setting('InternalEcpPublishingUrl'),
                    self.A_NAMESPACE.Setting('InternalOABUrl'),
                    self.A_NAMESPACE.Setting('InternalUMUrl'),
                    self.A_NAMESPACE.Setting('InternalWebClientUrls'),
                    self.A_NAMESPACE.Setting('PublicFolderServer'),
                    self.A_NAMESPACE.Setting('ExternalMailboxServer'),
                    self.A_NAMESPACE.Setting('ExternalMailboxServerRequiresSSL'),
                    self.A_NAMESPACE.Setting('ExternalMailboxServerAuthenticationMethods'),
                    self.A_NAMESPACE.Setting('EcpVoicemailUrlFragment'),
                    self.A_NAMESPACE.Setting('EcpEmailSubscriptionsUrlFragment'),
                    self.A_NAMESPACE.Setting('EcpTextMessagingUrlFragment'),
                    self.A_NAMESPACE.Setting('EcpDeliveryReportUrlFragment'),
                    self.A_NAMESPACE.Setting('EcpRetentionPolicyTagsUrlFragment'),
                    self.A_NAMESPACE.Setting('ExternalEcpUrl'),
                    self.A_NAMESPACE.Setting('EcpPublishingUrlFragment'),
                    self.A_NAMESPACE.Setting('ExternalEcpVoicemailUrl'),
                    self.A_NAMESPACE.Setting('ExternalEcpEmailSubscriptionsUrl'),
                    self.A_NAMESPACE.Setting('ExternalEcpTextMessagingUrl'),
                    self.A_NAMESPACE.Setting('ExternalEcpDeliveryReportUrl'),
                    self.A_NAMESPACE.Setting('EcpEmailSubscriptionsUrlFragment'),
                    self.A_NAMESPACE.Setting('ExternalEcpRetentionPolicyTagsUrl'),
                    self.A_NAMESPACE.Setting('ExternalEcpPublishingUrl'),
                    self.A_NAMESPACE.Setting('ExternalOABUrl'),
                    self.A_NAMESPACE.Setting('ExternalUMUrl'),
                    self.A_NAMESPACE.Setting('ExternalWebClientUrls'),
                    self.A_NAMESPACE.Setting('CrossOrganizationSharingEnabled'),
                    self.A_NAMESPACE.Setting('AlternateMailboxes'),
                    self.A_NAMESPACE.Setting('CasVersion'),
                    self.A_NAMESPACE.Setting('InternalPop3Connections'),
                    self.A_NAMESPACE.Setting('ExternalPop3Connections'),
                    self.A_NAMESPACE.Setting('InternalImap4Connections'),
                    self.A_NAMESPACE.Setting('ExternalImap4Connections'),
                    self.A_NAMESPACE.Setting('InternalSmtpConnections'),
                    self.A_NAMESPACE.Setting('ExternalSmtpConnections'),
                    self.A_NAMESPACE.Setting('InternalServerExclusiveConnect'),
                    self.A_NAMESPACE.Setting('ExternalServerExclusiveConnect'),
                    self.A_NAMESPACE.Setting('ExchangeRpcUrl'),
                    self.A_NAMESPACE.Setting('ShowGalAsDefaultView'),
                    self.A_NAMESPACE.Setting('AutoDiscoverSMTPAddress'),
                    self.A_NAMESPACE.Setting('InteropExternalEwsUrl'),
                    self.A_NAMESPACE.Setting('ExternalEwsVersion'),
                    self.A_NAMESPACE.Setting('InteropExternalEwsVersion'),
                    self.A_NAMESPACE.Setting('MobileMailboxPolicyInterop'),
                    self.A_NAMESPACE.Setting('GroupingInformation'),
                    self.A_NAMESPACE.Setting('UserMSOnline'),
                    self.A_NAMESPACE.Setting('MapiHttpEnabled')
                )
            ),
        )



from oauthlib.oauth2.rfc6749.errors import InvalidGrantError
import requests
from urllib.parse import urlparse, parse_qsl, urlencode, urlunparse
from oauthlib.oauth2 import LegacyApplicationClient
from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session


class OAuth2Connector:
    """OAuth2Connector is the main authentication mechanism for Microsoft Graph OAuth2 Authentication
    """
    AUTH_MAP = {
        'v1': {
            'authorize_url': 'https://login.microsoftonline.com/{tenant_id}/oauth2/authorize',
            'token_url': 'https://login.microsoftonline.com/{tenant_id}/oauth2/token',
            'resource': 'https://outlook.office365.com' 
        },
        'v2': {
            'authorize_url': 'https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/authorize',
            'token_url': 'https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token',
            'scope': 'https://outlook.office365.com/EWS.AccessAsUser.All'
        }
    }

    def __init__(self, endpoint_version='v1'):
        """OAuth2Connector is the base (parent) class of both Search and Delete classes.  It is used to perform either delegated authentication flows
        like: (Single-Page, Web Apps, Mobile & Native Apps - Grant Auth Flow) or you can use it in the application authentication auth flows like: (Client Credentials Grant Auth Flow)

        
        Args:
            client_id (str): Your Azure AD Application client ID
            client_secret (str): Your Azure AD Application client secret
            tenant_id (str): Your Azure AD tenant ID
            username (str, optional): A username used to authenticate to Azure or Office 365. Defaults to None. If provided, will use delegated authentication flows
            password (str, optional): The password used to authenticate to Azure or Office 365. Defaults to None. If provided, will use delegated authentication flows
            scopes (list, optional): A list of scopes defined during your Azure AD application registration. Defaults to ['https://graph.microsoft.com/.default'].
            verify_ssl (bool, optional): Whether to verify SSL or not. Defaults to True.
        """
        from .authentication import Authentication
        self.endpoint_version = endpoint_version
        self.verify = True
        self.username = Authentication.credentials[0]
        self.password = Authentication.credentials[1]
        self.client_id = Authentication.client_id
        self.client_secret = Authentication.client_secret
        self.tenant_id = Authentication.tenant_id
        self.access_token = Authentication.access_token
        self.redirect_uri = Authentication.redirect_uri
        self.authorize_url = self.AUTH_MAP.get(endpoint_version).get('authorize_url').format(tenant_id=self.tenant_id)
        self.token_url = self.AUTH_MAP.get(endpoint_version).get('token_url').format(tenant_id=self.tenant_id)
        if endpoint_version == 'v1':
            self.resource = self.AUTH_MAP.get(endpoint_version).get('resource')
        else:
            self.resource = None
        if not Authentication.oauth2_scope:
            if endpoint_version == 'v2':
                self.scope = [self.AUTH_MAP.get(endpoint_version).get('scope')]
            else:
                self.scope = None
        else:
            self.scope = Authentication.oauth2_scope
        self.session = requests.Session()
        self.session.headers = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        self.session.verify = self.verify
        self.expiration = None

    def __prompt_user(self, url, full_response=False):
        print('Please go here and authorize: ', url)
        response = input('Paste the full redirect URL here:')
        if full_response:
            return response
        if 'code=' in response:
            return response.split('code=')[-1].split('&')[0]
        elif 'id_token=' in response:
            return response.split('id_token=')[-1].split('&')[0]

    def __build_query_param_url(self, url, params):
        url_parse = urlparse(url)
        query = url_parse.query
        url_dict = dict(parse_qsl(query))
        url_dict.update(params)
        url_new_query = urlencode(url_dict)
        url_parse = url_parse._replace(query=url_new_query)
        return urlunparse(url_parse)

    def auth_code_grant(self):
        """Authorization Code Flow Grant
        Reference: https://docs.microsoft.com/en-us/azure/active-directory/develop/v2-oauth2-auth-code-flow
        """
        param = {
                'response_type': 'code',
                'client_id': self.client_id,
                'redirect_uri': self.redirect_uri,
                'state': '1234'
            }
        url = self.__build_query_param_url(self.authorize_url, param)
        authorization_code = self.__prompt_user(url)
        body = f'''
grant_type=authorization_code
&code={authorization_code}
&client_id={self.client_id}
&client_secret={self.client_secret}
&scope={self.scope}
&redirect_uri={self.redirect_uri} 
'''
        response = self.session.request('POST', self.token_url, data=body)
        return response.json().get('access_token')

    def client_credentials_grant(self):
        """Client Credentials Code Flow Grant
        Reference: https://docs.microsoft.com/en-us/azure/active-directory/develop/v2-oauth2-client-creds-grant-flow
        """
        if self.resource:
            body = {
                'resource' : self.resource,
                'client_id' : self.client_id,
                'client_secret' : self.client_secret,
                'grant_type' : 'client_credentials'
            }
        else:
            body = {
                'scope' : self.scope,
                'client_id' : self.client_id,
                'client_secret' : self.client_secret,
                'grant_type' : 'client_credentials'
            }
        response = self.session.request('POST', self.token_url, data=body).json()
        return response['access_token']

    def implicit_grant_flow(self):
        """Implicit Grant Flow
        Reference: https://docs.microsoft.com/en-us/azure/active-directory/develop/v2-oauth2-implicit-grant-flow
        """
        params = {
            'client_id': self.client_id,
            'response_type': 'id_token',
            'redirect_uri': self.redirect_uri,
            'scope': 'openid',
            'response_mode': 'fragment',
            'state': 1234,
            'nonce': 678910
        }
        url = self.__build_query_param_url(self.authorize_url, params)
        return self.__prompt_user(url)

    def web_application_flow(self):
        oauth = OAuth2Session(
            client_id=self.client_id,
            redirect_uri=self.redirect_uri,
            scope=self.scope
            )
        authorization_url, state = oauth.authorization_url(self.authorize_url)
        token = oauth.fetch_token(
            self.authority_url,
            client_secret=self.client_secret,
            authorization_response=self.__prompt_user(authorization_url, full_response=True)
        )
        return token

    def legacy_app_flow(self):
        """Resource Ownwer Password Credentials Grant Flow
        Reference: https://docs.microsoft.com/en-us/azure/active-directory/develop/v2-oauth-ropc
        """
        oauth = OAuth2Session(client=LegacyApplicationClient(client_id=self.client_id))
        try:
            if self.endpoint_version == 'v1':
                token = oauth.fetch_token(
                    token_url=self.token_url, 
                    username=self.username,
                    password=self.password, 
                    client_id=self.client_id,
                    client_secret=self.client_secret, 
                    resource=self.resource,
                    verify=self.verify
                )
            else:
                token = oauth.fetch_token(
                    token_url=self.token_url, 
                    username=self.username,
                    password=self.password, 
                    client_id=self.client_id,
                    client_secret=self.client_secret, 
                    scope=self.scope,
                    verify=self.verify
                )
        except InvalidGrantError as e:
            print(e)
            raise Exception('Please use another authorization method. I suggest trying the auth_code_grant() method.') 
        return token['access_token']

    def backend_app_flow(self):
        client = BackendApplicationClient(client_id=self.client_id)
        oauth = OAuth2Session(client=client)
        if self.endpoint_version == 'v1':
            token = oauth.fetch_token(
                token_url=self.token_url, 
                client_id=self.client_id, 
                client_secret=self.client_secret,
                resource=self.resource
            )
        else:
            token = oauth.fetch_token(
                token_url=self.token_url, 
                client_id=self.client_id, 
                client_secret=self.client_secret,
                scope=self.scope
            )
        return token['access_token']



from ..service import Operation
from ..utils.exceptions import UknownValueError
from ..utils.attributes import FOLDER_LIST, MESSAGE_ELEMENTS


class CreateItem(Operation):
    """CreateItem EWS Operation is used to create e-mail messages
    """

    RESULTS_KEY = 'Items'
    MESSAGE_DISPOSITION = [
        'SaveOnly',
        'SendOnly',
        'SendAndSaveCopy'
    ]
    BODY_TYPES = [
        'Best',
        'HTML',
        'Text'
    ]

    __MESSSAGE_ELEMENTS = []

    def __init__(self, type='message', message_disposition='SendAndSaveCopy', save_item_to='sentitems', **kwargs):
        """Creates an e-mail message

        Args:
            type (str, optional): The type of item to create. Defaults to 'message'.
            message_disposition (str, optional): The action to take when the item is created. Defaults to 'SendAndSaveCopy'.
            save_item_to (str, optional): Where to save the created e-mail message. Defaults to 'sentitems'.
        """
        self.__process_kwargs(kwargs)
        self.type = type
        if message_disposition not in self.MESSAGE_DISPOSITION:
            UknownValueError(provided_value=message_disposition, known_values=self.MESSAGE_DISPOSITION)
        self.message_disposition = message_disposition
        if save_item_to not in FOLDER_LIST:
            self.__save_item_to = self.T_NAMESPACE.FolderId(Id=save_item_to)
        else:
            self.__save_item_to = self.T_NAMESPACE.DistinguishedFolderId(Id=save_item_to)

    def __create_message(self):
        return self.T_NAMESPACE.Message(
            *self.__MESSSAGE_ELEMENTS
        )

    def __process_kwargs(self, kwargs):
        for key,val in kwargs.items():
            for item in self._get_recursively(MESSAGE_ELEMENTS, key):
                if isinstance(item, list):
                    if val in item:
                        if key in ['ToRecipients', 'Sender', 'CcRecipients', 'BccRecipients']:
                            self.__MESSSAGE_ELEMENTS.append(self.T_NAMESPACE(key, self.T_NAMESPACE.Mailbox(self.T_NAMESPACE.EmailAddress(val))))
                        else:
                            self.__MESSSAGE_ELEMENTS.append(self.T_NAMESPACE(key, val))
                else:
                    if key in MESSAGE_ELEMENTS:
                        if key in ['ToRecipients', 'Sender', 'CcRecipients', 'BccRecipients']:
                            self.__MESSSAGE_ELEMENTS.append(self.T_NAMESPACE(key, self.T_NAMESPACE.Mailbox(self.T_NAMESPACE.EmailAddress(val))))
                        else:
                            self.__MESSSAGE_ELEMENTS.append(self.T_NAMESPACE(key, val))

    def soap(self):
        item = None
        if self.type == 'message':
            item = self.__create_message()
        return self.M_NAMESPACE.CreateItem(
            self.M_NAMESPACE.SavedItemFolderId(
                self.__save_item_to
            ),
            self.M_NAMESPACE.Items(
                item
            ),
            MessageDisposition=self.message_disposition
        )



from ..service import Operation
from ..utils.exceptions import UknownValueError
from ..utils.attributes import FOLDER_LIST, MESSAGE_ELEMENTS


class GetServiceConfiguration(Operation):
    """GetServiceConfiguration EWS Operation retrieves
    service configuration details
    """

    RESULTS_KEY = 'ServiceConfigurationResponseMessageType'
    CONFIGURATION_NAMES = [
        'MailTips',
        'UnifiedMessagingConfiguration',
        'ProtectionRules'
    ]

    def __init__(self, configuration_name=None, acting_as=None):
        """Retrieves service configuration details.
        Default will attempt to retrieve them all.

        Args:
            configuration_name (list, optional): The name of one or more configuration items. Defaults to 'MailTips','UnifiedMessagingConfiguration','ProtectionRules'.
            acting_as (str, optional): If provided, will attempt to make call using provided user. Defaults to None.

        Raises:
            UknownValueError: Unknown value was provided
        """
        if configuration_name and configuration_name not in self.CONFIGURATION_NAMES:
            UknownValueError(provided_value=configuration_name, known_values=self.CONFIGURATION_NAMES)
        if configuration_name:
            self.configuration_name = configuration_name if isinstance(configuration_name, list) else [configuration_name]
        else:
            self.configuration_name = self.CONFIGURATION_NAMES
        self.acting_as = acting_as

    def __get_configuration_name(self):
        return_list = []
        for item in self.configuration_name:
            return_list.append(self.M_NAMESPACE.ConfigurationName(item))
        return return_list

    def soap(self):
        if self.acting_as:
            return self.M_NAMESPACE.GetServiceConfiguration(
                self.M_NAMESPACE.ActingAs(
                    self.T_NAMESPACE.EmailAddress(self.acting_as),
                    self.T_NAMESPACE.RoutingType('SMTP')
                ),
                self.M_NAMESPACE.RequestedConfiguration(
                    *self.__get_configuration_name()
                )
            )
        else:
            return self.M_NAMESPACE.GetServiceConfiguration(
                self.M_NAMESPACE.RequestedConfiguration(
                    *self.__get_configuration_name()
                )
            )



from ..service import Operation


class ResolveNames(Operation):
    """ResolveNames EWS Operation attempts
    to resolve the authenticated or provided name
    """

    RESULTS_KEY = 'ResolutionSet'

    def __init__(self, user=None):
        """Resolves the authenticated or provided name

        Args:
            user (str, optional): A user to attempt to resolve. Defaults to None.
        """
        self.user = user

    def soap(self):
        if not self.user:
            self.user = self.credentials[0]
        return self.M_NAMESPACE.ResolveNames(
            self.M_NAMESPACE.UnresolvedEntry(self.user),
            **{'ReturnFullContactData':"true"}
        )



from ..service.autodiscover import Autodiscover, Authentication


class GetDomainSettings(Autodiscover):
    """GetFederationInformation EWS Autodiscover endpoint
    retrieves the authenticated users federation information
    """

    def __init__(self, domain=None):
        """Retrieves the domain settings for the authenticated or provided domain.

        Args:
            user (str, optional): A user to retrieve user settings for. Defaults to None.
        """
        self.domain = domain

    def soap(self):
        return self.A_NAMESPACE.GetDomainSettingsRequestMessage(
            self.A_NAMESPACE.Request(
                self.A_NAMESPACE.Domains(
                    self.A_NAMESPACE.Domain(Authentication.credentials[0].split('@')[-1])
                ),
                self.A_NAMESPACE.RequestedSettings(
                    self.A_NAMESPACE.Setting('InternalEwsUrl'),
                    self.A_NAMESPACE.Setting('ExternalEwsUrl'),
                   
                )
            )
        )



from ..service import Operation
from ..utils.exceptions import UknownValueError
from ..utils.attributes import FOLDER_LIST, TRAVERSAL_LIST


class FindFolder(Operation):
    """FindFolder EWS Operation attempts to find a folder.
    Specifically this class is focused on finding search folders.
    """

    RESULTS_KEY = 'Folders'
    BASE_SHAPES = [
        'IdOnly',
        'Default',
        'AllProperties'
    ]
    BODY_TYPES = [
        'Best',
        'HTML',
        'Text'
    ]
    SEARCH_FILTERS = [
        'Contains',
        'Excludes',
        'Exists',
        'IsEqualTo',
        'IsNotEqualTo',
        'IsGreaterThan',
        'IsGreaterThanOrEqualTo',
        'IsLessThan',
        'IsLessThanOrEqualTo'
    ]

    def __init__(self, folder_id='searchfolders', traversal='Shallow'):
        """
        """
        self.folder_id = folder_id
        if traversal not in TRAVERSAL_LIST:
            raise UknownValueError(provided_value=traversal, known_values=TRAVERSAL_LIST)
        self.traversal = traversal

    def soap(self):
        return self.M_NAMESPACE.FindFolder(
            self.M_NAMESPACE.FolderShape(
                self.T_NAMESPACE.BaseShape('AllProperties'),
                self.T_NAMESPACE.AdditionalProperties(
                    self.T_NAMESPACE.FieldURI(FieldURI='folder:DisplayName')
                )
            ),
            self.M_NAMESPACE.IndexedPageFolderView(
                MaxEntriesReturned="10", Offset="0", BasePoint="Beginning"
            ),
            self.M_NAMESPACE.ParentFolderIds(
                self.T_NAMESPACE.DistinguishedFolderId(Id="searchfolders")
            )
        )



from ..service import Operation
from ..utils.exceptions import UknownValueError


class ConvertId(Operation):
    """ConvertId EWS Operation converts item and folder
    identifiers between formats.
    """

    RESULTS_KEY = '@Id'
    ID_FORMATS = [
        'EntryId',
        'EwsId',
        'EwsLegacyId',
        'HexEntryId',
        'OwaId',
        'StoreId'
    ]

    def __init__(self, user, item_id, id_type, convert_to):
        """Takes a specific user, item_id, id_type, and the 
        desired format to conver to as inputs.

        Args:
            user (str): The mailbox that the ID is associated with
            item_id (str): The item ID 
            id_type (str): The Item ID type
            convert_to (str): The format to conver the Item ID to

        Raises:
            UknownValueError: One or more provided values is unknown
        """
        self.user  = user
        self.item_id = item_id
        if id_type not in self.ID_FORMATS:
            UknownValueError(provided_value=id_type, known_values=self.ID_FORMATS)
        if  convert_to not in self.ID_FORMATS:
            UknownValueError(provided_value=convert_to, known_values=self.ID_FORMATS)
        self.id_type = id_type
        self.convert_to = convert_to

    def soap(self):
        return self.M_NAMESPACE.ConvertId(
            self.M_NAMESPACE.SourceIds(
                self.T_NAMESPACE.AlternateId(
                    Format=self.id_type, Id=self.item_id, Mailbox=self.user
                )
            ),
            DestinationFormat=self.convert_to
        )



from ..service import Operation
from lxml import etree
from ..utils.exceptions import UknownValueError
from ..utils.attributes import FOLDER_LIST, FIELD_URI_MAP, TRAVERSAL_LIST
from ..utils.searchfilter import SearchFilter


class CreateFolder(Operation):
    """FindItem EWS Operation retrieves details about an item.
    """

    RESULTS_KEY = 'CreateFolderResponseMessage'
    BASE_SHAPES = [
        'IdOnly',
        'Default',
        'AllProperties'
    ]
    BODY_TYPES = [
        'Best',
        'HTML',
        'Text'
    ]
    SEARCH_FILTERS = [
        'Contains',
        'Excludes',
        'Exists',
        'IsEqualTo',
        'IsNotEqualTo',
        'IsGreaterThan',
        'IsGreaterThanOrEqualTo',
        'IsLessThan',
        'IsLessThanOrEqualTo'
    ]

    def __init__(self, search_string, search_folder=True, display_name='Search Folder', search_all_folders=True, base_folder='inbox', traversal='Deep'):
        """Creates a Folder. Default behavior is to create a Search Folder (eDiscovery Search Folder)

            search_string:
                Detailed Information:
                Each search filter is made of several elements. These are:
                    Field Name - Retrieve all available Field Names by using SearchFilter.FIELD_NAMES
                    Expression - Retrieve all expressions by using SearchFilter.EXPRESSIONS
                    Value - Values are typically strings which make sens for field names
                A logic operator within the SearchFilter class is one of:
                    and - This will perform a boolean AND search operation between two or more search expressions
                    or - This will perform a logical OR search operation between two or more search expressions
                    not - This will perform a search expression that negates a search expression
                Examples:
                    SearchFilter('Body contains Hello World')
                    SearchFilter('Subject contains phishing and IsRead IsEqualTo true')
                    SearchFilter('Subject contains phishing and IsRead IsEqualTo true or Subject contains microsoft', to_string=True)

        Args:
            search_string (str): A search string which is converted to a Search Filter
            search_folder (bool, optional): Creates a Search Folder. Defaults to True.
            display_name (str, optional): The display name of the search folder. Defaults to 'Search Folder'.
            base_folder (str, optional): The base folder the search folder should look at. Defaults to 'inbox'.
            traversal (str, optional): The traversal type. Options are Deep and Shallow. Defaults to 'Deep'.
        """
        self.search_string = SearchFilter(search_string)
        self.display_name = display_name
        if search_folder:
            self.folder_id = [self.T_NAMESPACE.DistinguishedFolderId(Id='searchfolders')]
        else:
            self.folder_id = self.__get_distinguished_folder_ids()
        #if search_all_folders:
        #    self.base_folder_ids = self.__get_distinguished_folder_ids()
        #else:
        self.base_folder_ids = [self.T_NAMESPACE.DistinguishedFolderId(Id=base_folder)]
        if traversal not in TRAVERSAL_LIST:
            raise UknownValueError(provided_value=traversal, known_values=TRAVERSAL_LIST)
        self.traversal = traversal

    def __get_distinguished_folder_ids(self):
        return_list = []
        for item in FOLDER_LIST:
            return_list.append(self.T_NAMESPACE.DistinguishedFolderId(Id=item))
        return return_list

    def soap(self):
        return self.M_NAMESPACE.CreateFolder(
            self.M_NAMESPACE.ParentFolderId(
                *self.folder_id
            ),
            self.M_NAMESPACE.Folders(
                self.T_NAMESPACE.SearchFolder(
                    self.T_NAMESPACE.DisplayName(self.display_name),
                    self.T_NAMESPACE.SearchParameters(
                        self.search_string,
                        self.T_NAMESPACE.BaseFolderIds(
                            *self.base_folder_ids
                        ),
                        Traversal=self.traversal
                    )
                )
            )
        )



from ..service import Operation


class SyncFolderHierarchy(Operation):
    """SyncFolderHierarchy EWS Operation will retrieve 
    a mailboxes folder hierarchy
    """

    RESULTS_KEY = 'Changes'
    FOLDER_LIST = [
        'msgfolderroot',
        'calendar',
        'contacts',
        'deleteditems',
        'drafts',
        'inbox',
        'journal',
        'notes',
        'outbox',
        'sentitems',
        'tasks',
        'junkemail',
        'searchfolders',
        'voicemail',
        'recoverableitemsdeletions',
        'recoverableitemsversions',
        'recoverableitemspurges',
        'recipientcache',
        'quickcontacts',
        'conversationhistory',
        'todosearch',
        'mycontacts',
        'imcontactlist',
        'peopleconnect',
        'favorites'
    ]

    def __init__(self, well_known_folder_name=None):
        """Retrieve the authenticated users mailbox folder hierarchy.

        Args:
            well_known_folder_name (str, optional): The well known folder name. Defaults to all known folder names.
        """
        self.attachment_id = well_known_folder_name

    def soap(self):
        folder_id_list = []
        for folder in self.FOLDER_LIST:
            folder_id_list.append(self.T_NAMESPACE.DistinguishedFolderId(Id=folder))
        return self.M_NAMESPACE.SyncFolderHierarchy(
            self.M_NAMESPACE.FolderShape(
                self.T_NAMESPACE.BaseShape('AllProperties')
            ),
            self.M_NAMESPACE.SyncFolderId(
                *folder_id_list
            )
        )



from ..service import Operation
from ..utils.exceptions import UknownValueError


class SearchMailboxes(Operation):
    """SearchMailboxes EWS Operation will search using a query on one or more provided
    reference id's
    """

    RESULTS_KEY = 'SearchPreviewItem'
    SEARCH_SCOPES = ['All', 'PrimaryOnly', 'ArchiveOnly']

    def __init__(self, query, reference_id, search_scope='All'):
        """Searches one or more reference id's using the provided query and search_scope.

        Args:
            query (str): The Advanced Query Syntax (AQS) to search with.
            reference_id (list): One or more mailbox reference Id's
            search_scope (str, optional): The search scope. Defaults to 'All'.

        Raises:
            UknownValueError: The provided search scope is unknown.
        """
        self.query = query
        if not isinstance(reference_id, list):
            reference_id = [reference_id]
        self.reference_id = reference_id
        if search_scope not in self.SEARCH_SCOPES:
            raise UknownValueError('The search_scope ({}) you provided is not one of {}'.format(search_scope, ','.join([x for x in self.SEARCH_SCOPES])))
        if search_scope in self.SEARCH_SCOPES:
            self.scope = search_scope

    def soap(self):
        return self.M_NAMESPACE.SearchMailboxes(
            self.M_NAMESPACE.SearchQueries(
                self.T_NAMESPACE.MailboxQuery(
                    self.T_NAMESPACE.Query(self.query),
                    self.T_NAMESPACE.MailboxSearchScopes(*self.__get_search_scope())
                )
            ),
            self.M_NAMESPACE.ResultType('PreviewOnly')
        )

    def __get_search_scope(self):
        mailbox_soap_element = []
        for item in self.reference_id:
            mailbox_soap_element.append(
                self.T_NAMESPACE.MailboxSearchScope(
                    self.T_NAMESPACE.Mailbox(item),
                    self.T_NAMESPACE.SearchScope(self.scope)
                )
            )
        return mailbox_soap_element



class UknownValueError(ValueError):
    """Raised when the provided value is unkown or is not 
    in a specified list or dictionary map
    """
    def __init__(self, provided_value=None, known_values=None):
        if provided_value and known_values:
            if isinstance(known_values, list):
                super().__init__("The provided value {} is unknown. Please provide one of the following values: '{}'".format(
                    provided_value,
                    ','.join([x for x in known_values])
                ))
        else:
            pass


import abc
import requests
from lxml.builder import ElementMaker
from lxml import etree
from bs4 import BeautifulSoup

from ..core import Core, Authentication
from ..utils.attributes import RESPONSE_CODES


class Base(Core):
    """The Base class is used by all endpoints and network
    communications. It defines the base structure of all namespaces.
    It is inherited by the Autodiscover & Operation classes
    """

    SOAP_REQUEST_HEADER = {'content-type': 'text/xml', 'charset': 'UTF-8'}
    NAMESPACE_MAP = {
        'soap': "http://schemas.xmlsoap.org/soap/envelope/", 
        'm': "http://schemas.microsoft.com/exchange/services/2006/messages", 
        't': "http://schemas.microsoft.com/exchange/services/2006/types", 
        'a': "http://schemas.microsoft.com/exchange/2010/Autodiscover",
    }
    SOAP_MESSAGE_ELEMENT = ElementMaker(
        namespace=NAMESPACE_MAP['soap'],
        nsmap={
            'soap': NAMESPACE_MAP['soap'], 
            'm': NAMESPACE_MAP['m'], 
            't': NAMESPACE_MAP['t']
        }
    )
    SOAP_NAMESPACE = ElementMaker(namespace=NAMESPACE_MAP['soap'],nsmap={'soap': "http://schemas.xmlsoap.org/soap/envelope/"})
    M_NAMESPACE =  ElementMaker(namespace=NAMESPACE_MAP['m'],nsmap={'m': "http://schemas.microsoft.com/exchange/services/2006/messages"})
    T_NAMESPACE = ElementMaker(namespace=NAMESPACE_MAP['t'], nsmap={'t': "http://schemas.microsoft.com/exchange/services/2006/types"})
    XML_ENVELOPE = SOAP_MESSAGE_ELEMENT = ElementMaker(
        namespace=NAMESPACE_MAP['soap'],
        nsmap={
            'soap': NAMESPACE_MAP['soap'], 
            'm': NAMESPACE_MAP['m'], 
            't': NAMESPACE_MAP['t'],
            'xsi': "http://www.w3.org/2001/XMLSchema-instance",
            'xs': "http://www.w3.org/2001/XMLSchema"
        }
    )
    NULL_ELEMENT = ElementMaker()
    HEADER_ELEMENT = ElementMaker(
        namespace=NAMESPACE_MAP['soap'],
        nsmap={
            'soap': NAMESPACE_MAP['soap'], 
            't': NAMESPACE_MAP['t']
        }
    )
    BODY_ELEMENT = SOAP_MESSAGE_ELEMENT.Body
    EXCHANGE_VERSION_ELEMENT = T_NAMESPACE.RequestServerVersion

    @property
    def raw_xml(self):
        return self.__raw_xml

    @raw_xml.setter
    def raw_xml(self, value):
        self.__raw_xml = value

    @abc.abstractmethod
    def get(self):
        raise NotImplementedError

    def _impersonation_header(self):
        if hasattr(Authentication, 'impersonate_as') and Authentication.impersonate_as:
            return self.T_NAMESPACE.ExchangeImpersonation(
                self.T_NAMESPACE.ConnectingSID(
                    self.T_NAMESPACE.PrimarySmtpAddress(Authentication.impersonate_as)
                )
            )
        return ''

    def __parse_convert_id_error_message(self, error_message):
        result = error_message.split('Please use the ConvertId method to convert the Id from ')[1].split(' format.')[0]
        return result.split(' to ')

    def __process_response(self, response):
        self.__logger.debug('SOAP REQUEST: {}'.format(response))
        self.raw_xml = response
        namespace = self.NAMESPACE_MAP
        namespace_dict = {}
        for key,val in namespace.items():
            namespace_dict[val] = None
        return self.parse_response(response, namespace_dict=namespace_dict)

    def __log_warning(self, kwargs):
        log_message = f'''
The server responded with:
    response code: {kwargs.get('response_code')}
    error code: {kwargs.get('error_code')}
    from: {kwargs.get('from')}
    error message: {kwargs.get('message_text')}
'''
        if kwargs.get('additional_text'):
            log_message += f'''\n\t{kwargs.get('additional_text')}'''
        return log_message

    def run(self):
        """The Base class run method is used for all SOAP requests for
        every endpoint defined

        Returns:
            BeautifulSoup: Returns a BeautifulSoup object or None.
        """
        for item in Authentication.__dict__.keys():
            if not item.startswith('_'):
                if hasattr(Authentication, item):
                    setattr(self, item, getattr(Authentication, item))

        for version in Authentication.exchange_versions:
            for endpoint in Authentication.ews_url:
                if self.__class__.__base__.__name__ == 'Operation' and 'autodiscover' in endpoint:
                    self.__logger.debug('{} == Operation so skipping endpoint {}'.format(self.__class__.__base__.__name__, endpoint))
                    continue
                elif self.__class__.__base__.__name__ == 'Autodiscover' and 'autodiscover' not in endpoint:
                    self.__logger.debug('{} == Autodiscover so skipping endpoint {}'.format(self.__class__.__base__.__name__, endpoint))
                    continue
                try:
                    self.__logger.info('Sending SOAP request to {}'.format(endpoint))
                    self.__logger.info('Setting Exchange Version header to {}'.format(version))
                    body = self.get(version).decode("utf-8")
                    self.__logger.debug('EWS SOAP Request Body: {}'.format(body))
                    if Authentication.auth_header:
                        header_dict = Authentication.auth_header
                        header_dict.update(self.SOAP_REQUEST_HEADER)
                    else:
                        header_dict = self.SOAP_REQUEST_HEADER
                    self.__logger.debug(f"Headers: {header_dict}")
                    response = requests.post(
                        url=endpoint,
                        data=body,
                        headers=header_dict,
                        auth=Authentication.credentials,
                        verify=True
                    )

                    self.__logger.debug('Response HTTP status code: {}'.format(response.status_code))
                    self.__logger.debug('Response text: {}'.format(response.text))
                    self.__logger.debug('Response content: {}'.format(response.content))

                    parsed_response = BeautifulSoup(response.content, 'xml')
                    if not parsed_response.contents:
                        self.__logger.warning(
                            'The server responded with empty content to POST-request '
                            'from {current}'.format(current=self.__class__.__name__))

                    response_code = getattr(parsed_response.find('ResponseCode'), 'string', None)
                    error_code = getattr(parsed_response.find('ErrorCode'), 'string', None)
                    message_text = getattr(parsed_response.find('MessageText'), 'string', None)
                    error_message = getattr(parsed_response.find('ErrorMessage'), 'string', None)
                    fault_message = getattr(parsed_response.find('faultcode'), 'string', None)
                    fault_string = getattr(parsed_response.find('faultstring'), 'string', None)

                    if 'NoError' in (response_code, error_code):
                        return self.__process_response(parsed_response)
                    if response_code in RESPONSE_CODES or error_code in RESPONSE_CODES or error_message:
                        warning_dict = {
                            'response_code': response_code,
                            'error_code': error_code,
                            'from': self.__class__.__name__,
                            'message_text': message_text
                        }
                        if 'ConvertId' in message_text:
                            return self.__parse_convert_id_error_message(message_text)
                        if 'ErrorAccessDenied' in (response_code, error_code) and endpoint in ('GetSearchableMailboxes', 'SearchMailboxes'):
                            warning_dict.update({
                                'additional_text': 'Please make sure you have Discovery Management rights: https://docs.microsoft.com/en-us/Exchange/policy-and-compliance/ediscovery/assign-permissions?redirectedfrom=MSDN&view=exchserver-2019'
                            })
                        elif error_message:
                            warning_dict.update({
                                'additional_text': error_message
                            })
                        self.__logger.info(self.__log_warning(warning_dict))
                        continue
                    elif fault_message or fault_string:
                        warning_dict = {
                            'response_code': response_code,
                            'error_code': error_code,
                            'from': self.__class__.__name__,
                            'message_text': fault_message,
                            'additional_text': fault_string
                        }
                        self.__logger.info(self.__log_warning(warning_dict))
                        continue
                    else:
                        warning_dict = {
                            'response_code': 'Unknown',
                            'error_code': 'Unknown',
                            'from': self.__class__.__name__,
                            'message_text': message_text,
                            'additional_text': 'This error is unrecognized and is not a valid response code or error code.'
                        }
                        self.__logger.info(self.__log_warning(warning_dict))
                        continue
                except:
                    pass



from ..service import Operation
from ..utils.exceptions import UknownValueError
from ..utils.attributes import FOLDER_LIST, TRAVERSAL_LIST


class FindItem(Operation):
    """FindItem EWS Operation retrieves details about an item.
    """

    RESULTS_KEY = 'Message'
    BASE_SHAPES = [
        'IdOnly',
        'Default',
        'AllProperties'
    ]
    BODY_TYPES = [
        'Best',
        'HTML',
        'Text'
    ]

    def __init__(self, query_string, distinguished_folder_name='inbox', base_shape='AllProperties', include_mime_content=True, body_type='Best', traversal='Shallow', reset_cache=False, return_deleted_items=True, return_highlight_terms=True):
        """Retrieves results from a query string

        Args:
            item_id (str): The item id you want to get information about.
            change_key (str, optional): The change key of the item. Defaults to None.
            base_shape (str, optional): The base shape of the returned item. Defaults to 'AllProperties'.
            include_mime_content (bool, optional): Whether or not to include MIME content. Defaults to True.
            body_type (str, optional): The item body type. Defaults to 'Best'.
        """
        if distinguished_folder_name:
            self.folder_name = [self.T_NAMESPACE.DistinguishedFolderId(Id=distinguished_folder_name)]
        else:
            self.folder_name = self.__get_distinguished_folder_ids()
        self.query_string = query_string
        self.include_mime_content = include_mime_content
        if base_shape not in self.BASE_SHAPES:
            raise UknownValueError(provided_value=base_shape, known_values=self.BASE_SHAPES)
        self.base_shape = base_shape
        if body_type not in self.BODY_TYPES:
            raise UknownValueError(provided_value=body_type, known_values=self.BODY_TYPES)
        self.body_type = body_type
        if traversal not in TRAVERSAL_LIST:
            raise UknownValueError(provided_value=traversal, known_values=TRAVERSAL_LIST)
        self.traversal = traversal

        self.query_properties = {}
        if reset_cache:
            self.query_properties.update({
                'ResetCache': 'true'
            })
        if return_deleted_items:
            self.query_properties.update({
                'ReturnDeletedItems': 'true'
            })
        if return_highlight_terms:
            self.query_properties.update({
                'ReturnHighlightTerms': 'true'
            })

    def __get_distinguished_folder_ids(self):
        return_list = []
        for item in FOLDER_LIST:
            return_list.append(self.T_NAMESPACE.DistinguishedFolderId(Id=item))
        return return_list

    def soap(self):
        return self.M_NAMESPACE.FindItem(
            self.M_NAMESPACE.ItemShape(
                self.M_NAMESPACE.BaseShape(self.base_shape),
                self.M_NAMESPACE.IncludeMimeContent(str(self.include_mime_content).lower()),
                self.M_NAMESPACE.BodyType(self.body_type),
                self.M_NAMESPACE.FilterHtmlContent('false')
            ),
            self.M_NAMESPACE.ParentFolderIds(
                *self.folder_name
            ),
            self.M_NAMESPACE.QueryString(self.query_string, **self.query_properties),
            Traversal=self.traversal
        )
